import warnings
from typing import List, Tuple, Union

import ase.atoms
from structuretoolkit.build import pyxtal as _pyxtal

from pyiron_atomistics.atomistics.structure.atoms import Atoms, ase_to_pyiron
from pyiron_atomistics.atomistics.structure.structurestorage import StructureStorage

publication = {
    "pyxtal": {
        "title": "PyXtal: A Python library for crystal structure generation and symmetry analysis",
        "journal": "Computer Physics Communications",
        "volume": "261",
        "pages": "107810",
        "year": "2021",
        "issn": "0010-4655",
        "doi": "https://doi.org/10.1016/j.cpc.2020.107810",
        "url": "http://www.sciencedirect.com/science/article/pii/S0010465520304057",
        "author": "Scott Fredericks and Kevin Parrish and Dean Sayre and Qiang Zhu",
    }
}


def pyxtal(
    group: Union[int, List[int]],
    species: Tuple[str],
    num_ions: Tuple[int],
    dim=3,
    repeat=1,
    storage=None,
    allow_exceptions=True,
    **kwargs,
) -> Union[Atoms, StructureStorage]:
    """
    Generate random crystal structures with PyXtal.

    `group` must be between 1 and the largest possible value for the given dimensionality:
        dim=3 => 1 - 230 (space groups)
        dim=2 => 1 -  80 (layer groups)
        dim=1 => 1 -  75 (rod groups)
        dim=0 => 1 -  58 (point groups)

    When `group` is passed as a list of integers or `repeat>1`, generate multiple structures and return them in a :class:`.StructureStorage`.

    Args:
        group (list of int, or int): the symmetry group to generate or a list of them
        species (tuple of str): which species to include, defines the stoichiometry together with `num_ions`
        num_ions (tuple of int): how many of each species to include, defines the stoichiometry together with `species`
        dim (int): dimensionality of the symmetry group, 0 is point groups, 1 is rod groups, 2 is layer groups and 3 is space groups
        repeat (int): how many random structures to generate
        storage (:class:`.StructureStorage`, optional): when generating multiple structures, add them to this instead of creating a new storage
        allow_exceptions (bool): when generating multiple structures, silence errors when the requested stoichiometry and symmetry group are incompatible
        **kwargs: passed to `pyxtal.pyxtal` function verbatim

    Returns:
        :class:`~.Atoms`: the generated structure, if repeat==1 and only one symmetry group is requested
        :class:`.StructureStorage`: a storage of all generated structure, if repeat>1 or multiple symmetry groups are requested

    Raises:
        ValueError: if `species` and `num_ions` are not of the same length
        ValueError: if stoichiometry and symmetry group are incompatible and allow_exceptions==False or only one structure is requested
    """
    ret = _pyxtal(
        group=group,
        species=species,
        num_ions=num_ions,
        dim=dim,
        repeat=repeat,
        allow_exceptions=allow_exceptions,
        **kwargs,
    )
    if isinstance(ret, ase.atoms.Atoms):
        return ase_to_pyiron(ret)
    else:
        stoich = "".join(f"{s}{n}" for s, n in zip(species, num_ions))
        if storage is None:
            storage = StructureStorage()
        for struct in ret:
            storage.add_structure(
                struct["atoms"],
                identifier=f"{stoich}_{struct['symmetry']}_{struct['repeat']}",
                symmetry=struct["symmetry"],
                repeat=struct["repeat"],
            )
        return storage



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "GPLv3", see the LICENSE file.

from collections import OrderedDict

import numpy as np
import scipy.constants
import spglib
from atomistics.workflows.elastic.elastic_moduli import ElasticProperties
from atomistics.workflows.elastic.helper import (
    analyse_structures_helper,
    generate_structures_helper,
)
from pyiron_base import JobGenerator

from pyiron_atomistics.atomistics.master.parallel import AtomisticParallelMaster

__author__ = "Yury Lysogorskiy"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"


class ElasticMatrixCalculator(object):
    def __init__(
        self, basis_ref, num_of_point=5, eps_range=0.005, sqrt_eta=True, fit_order=2
    ):
        self.basis_ref = basis_ref.copy()
        self.num_of_point = num_of_point
        self.eps_range = eps_range
        self.sqrt_eta = sqrt_eta
        self.fit_order = fit_order
        self._data = OrderedDict()
        self._structure_dict = OrderedDict()
        self.SGN = None
        self.v0 = None
        self.LC = None
        self.Lag_strain_list = []
        self.epss = np.array([])
        self.zero_strain_job_name = "s_e_0"

    def generate_structures(self):
        """

        Returns:

        """
        self._data, self._structure_dict = generate_structures_helper(
            structure=self.basis_ref.copy(),
            eps_range=self.eps_range,
            num_of_point=self.num_of_point,
            zero_strain_job_name=self.zero_strain_job_name,
            sqrt_eta=self.sqrt_eta,
        )

        # Copy self._data to properties for backwards compatibility
        self.Lag_strain_list = self._data["Lag_strain_list"]
        self.epss = self._data["epss"]
        self.v0 = self._data["v0"]
        self.LC = self._data["LC"]
        self.SGN = self._data["SGN"]
        return self._structure_dict

    def analyse_structures(self, output_dict):
        """

        Returns:

        """
        sym_dict, elastic = analyse_structures_helper(
            output_dict=output_dict,
            sym_dict=self._data,
            fit_order=self.fit_order,
            zero_strain_job_name=self.zero_strain_job_name,
        )
        self._data.update(elastic)

    @staticmethod
    def subjob_name(i, eps):
        """

        Args:
            i:
            eps:

        Returns:

        """
        return ("s_%s_e_%.5f" % (i, eps)).replace(".", "_").replace("-", "m")


class ElasticJobGenerator(JobGenerator):
    @property
    def parameter_list(self):
        """

        Returns:
            (list)
        """
        return [
            [job_name, basis] for job_name, basis in self._job.structure_dict.items()
        ]

    @staticmethod
    def job_name(parameter):
        return str(parameter[0])

    def modify_job(self, job, parameter):
        job.structure = parameter[1]
        return job


class ElasticMatrixJob(AtomisticParallelMaster):
    def __init__(self, project, job_name="elasticmatrix"):
        super(ElasticMatrixJob, self).__init__(project, job_name)
        self.__name__ = "ElasticMatrixJob"
        self.__version__ = "0.0.1"
        self.input["num_of_points"] = (
            5,
            "number of sample point per deformation directions",
        )
        self.input["fit_order"] = (2, "order of the fit polynom")
        self.input["eps_range"] = (0.005, "strain variation")
        self.input["sqrt_eta"] = (
            True,
            "calculate self-consistently sqrt of stress matrix eta",
        )
        self._data = OrderedDict()
        self.structure_dict = OrderedDict()
        self.property_calculator = None
        self.hdf_storage_group = "elasticmatrix"
        self._job_generator = ElasticJobGenerator(master=self)

    def create_calculator(self):
        if self.property_calculator is None:
            self.property_calculator = ElasticMatrixCalculator(
                basis_ref=self.ref_job.structure.copy(),
                num_of_point=int(self.input["num_of_points"]),
                eps_range=self.input["eps_range"],
                sqrt_eta=self.input["sqrt_eta"],
                fit_order=int(self.input["fit_order"]),
            )
            self.structure_dict = self.property_calculator.generate_structures()
            self._data.update(self.property_calculator._data)

    def run_static(self):
        self.create_calculator()
        super(ElasticMatrixJob, self).run_static()

    def run_if_interactive(self):
        self.create_calculator()
        super(ElasticMatrixJob, self).run_if_interactive()

    def run_if_refresh(self):
        self.create_calculator()
        super(ElasticMatrixJob, self).run_if_refresh()

    def collect_output(self):
        if not self._data:
            self.from_hdf()
        self.create_calculator()

        energies = {}
        self._data["id"] = []
        if self.server.run_mode.interactive:
            child_id = self.child_ids[0]
            self._data["id"].append(child_id)
            child_job = self.project_hdf5.inspect(child_id)
            energies = {
                job_name: energy
                for job_name, energy in zip(
                    self.structure_dict.keys(), child_job["output/generic/energy_tot"]
                )
            }
        else:
            for job_id in self.child_ids:
                ham = self.project_hdf5.inspect(job_id)
                en = ham["output/generic/energy_tot"][-1]
                energies[ham.job_name] = en
                self._data["id"].append(ham.job_id)

        self.property_calculator.analyse_structures(energies)
        self._data.update(self.property_calculator._data)
        self.to_hdf()

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore object from hdf5 format
        :param hdf: Optional hdf5 file, otherwise self._hdf5 is used.
        :param group_name: Optional hdf5 group in the hdf5 file.
        """
        super(ElasticMatrixJob, self).from_hdf(hdf=hdf, group_name=group_name)
        try:
            with self.project_hdf5.open("output") as hdf5_out:
                self._data.update(hdf5_out[self.hdf_storage_group])
        except Exception as e:
            print(e)

    def to_hdf(self, hdf=None, group_name=None):
        super(ElasticMatrixJob, self).to_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("output") as hdf5_out:
            hdf5_out[self.hdf_storage_group] = self._data



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
from pyiron_base import DataContainer, FlattenedStorage
from pyiron_snippets.deprecate import deprecate

__author__ = "Joerg Neugebauer, Sam Waseda"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sam Waseda"
__email__ = "waseda@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class NeighborsTrajectory(DataContainer):
    """
    This class generates the neighbors for a given atomistic trajectory. The resulting indices, distances, and vectors
    are stored as numpy arrays.
    """

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        object.__setattr__(instance, "_has_structure", None)
        return instance

    def __init__(
        self,
        init=None,
        has_structure=None,
        num_neighbors=12,
        table_name="neighbors_traj",
        store=None,
        **kwargs,
    ):
        """

        Args:
            has_structure (:class:`.HasStructure`): object containing the structures to compute the neighbors on
            num_neighbors (int): The cutoff for the number of neighbors
            table_name (str): Table name for the base `DataContainer` (stores this object as a group in a HDF5 file with
                              this name)
            store (FlattenedStorage): internal storage that should be used to store the neighborhood information,
                                      creates a new one if not provided; if provided and not empty it must be compatible
                                      with the lengths of the structures in `has_structure`, but this is *not* checked
            **kwargs (dict): Additional arguments to be passed to the `get_neighbors()` routine
                             (eg. cutoff_radius, norm_order , etc.)
        """
        super().__init__(init=init, table_name=table_name)
        self._flat_store = store if store is not None else FlattenedStorage()
        self._flat_store.add_array(
            "indices", dtype=np.int64, shape=(num_neighbors,), per="element", fill=-1
        )
        self._flat_store.add_array(
            "distances", dtype=np.float64, shape=(num_neighbors,), per="element"
        )
        self._flat_store.add_array(
            "vecs", dtype=np.float64, shape=(num_neighbors, 3), per="element"
        )
        self._flat_store.add_array(
            "shells", dtype=np.int64, shape=(num_neighbors,), per="element"
        )
        self._num_neighbors = num_neighbors
        self._get_neighbors_kwargs = kwargs
        self.has_structure = has_structure

    @property
    def has_structure(self):
        return self._has_structure

    @has_structure.setter
    def has_structure(self, value):
        if value is not None:
            self._has_structure = value
            self._compute_neighbors()

    @property
    def indices(self):
        """
        Neighbour indices (excluding itself) of each atom computed using the get_neighbors_traj() method

        If the structures have different number of atoms, the array will have -1 on indices that are invalid.

        Returns:
            numpy.ndarray: An int array of dimension N_steps / stride x N_atoms x N_neighbors
        """
        return self._flat_store.get_array_filled("indices")

    @property
    def distances(self):
        """
        Neighbour distances (excluding itself) of each atom computed using the get_neighbors_traj() method

        If the structures have different number of atoms, the array will have NaN on indices that are invalid.

        Returns:
            numpy.ndarray: A float array of dimension N_steps / stride x N_atoms x N_neighbors
        """
        return self._flat_store.get_array_filled("distances")

    @property
    def vecs(self):
        """
        Neighbour vectors (excluding itself) of each atom computed using the get_neighbors_traj() method

        If the structures have different number of atoms, the array will have NaN on indices that are invalid.

        Returns:
            numpy.ndarray: A float array of dimension N_steps / stride x N_atoms x N_neighbors x 3
        """
        return self._flat_store.get_array_filled("vecs")

    @property
    def shells(self):
        """
        Neighbor shell indices (excluding itself) of each atom computed using the get_neighbors_traj() method.

        For trajectories with non constant amount of particles this array may contain -1 for invalid values, i.e.

        Returns:
            ndarray: An int array of dimension N_steps / stride x N_atoms x N_neighbors x 3
        """
        return self._flat_store.get_array_filled("shells")

    @property
    def num_neighbors(self):
        """
        The maximum number of neighbors to be computed

        Returns:

            int: The max number of neighbors
        """
        return self._num_neighbors

    def _compute_neighbors(self):
        for i, struct in enumerate(self._has_structure.iter_structures()):
            if (
                i < len(self._flat_store)
                and (self._flat_store["indices", i] != -1).all()
            ):
                # store already has valid entries for this structure, so skip it
                continue
            # Change the `allow_ragged` based on the changes in get_neighbors()
            neigh = struct.get_neighbors(
                num_neighbors=self._num_neighbors,
                allow_ragged=False,
                **self._get_neighbors_kwargs,
            )
            if i >= len(self._flat_store):
                self._flat_store.add_chunk(
                    len(struct),
                    indices=neigh.indices,
                    distances=neigh.distances,
                    vecs=neigh.vecs,
                    shells=neigh.shells,
                )
            else:
                self._flat_store.set_array("indices", i, neigh.indices)
                self._flat_store.set_array("distances", i, neigh.distances)
                self._flat_store.set_array("vecs", i, neigh.vecs)
                self._flat_store.set_array("shells", i, neigh.shells)
        return (
            self._flat_store.get_array_filled("indices"),
            self._flat_store.get_array_filled("distances"),
            self._flat_store.get_array_filled("vecs"),
        )

    @deprecate(
        "This has no effect, neighbors are automatically called on instantiation."
    )
    def compute_neighbors(self):
        """
        Compute the neighbors across the trajectory
        """
        pass



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os

import pandas
from pyiron_base import state
from pyiron_snippets.resources import ResourceResolver

from pyiron_atomistics.vasp.potential import VaspPotentialAbstract

__author__ = "Osamu Waseda"
__copyright__ = (
    "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Osamu Waseda"
__email__ = "waseda@mpie.de"
__status__ = "development"
__date__ = "Sep 20, 2019"


class SphinxJTHPotentialFile(VaspPotentialAbstract):
    """
    The Potential class is derived from the PotentialAbstract class, but instead of loading the potentials from a list,
    the potentials are loaded from a file.

    Args:
        xc (str): Exchange correlation functional ['PBE', 'LDA']
    """

    resource_plugin_name = "sphinx"

    @classmethod
    def _get_resolver(cls):
        env = os.environ
        return (
            super()
            ._get_resolver()
            .chain(
                ResourceResolver(
                    [env[var] for var in ("CONDA_PREFIX", "CONDA_DIR") if var in env],
                    "share",
                    "sphinxdft",
                )
            )
        )

    def __init__(self, xc=None, selected_atoms=None):
        potential_df = self._get_potential_df(
            file_name_lst={"potentials_sphinx.csv"},
        )
        if xc == "PBE":
            default_df = self._get_potential_default_df(
                file_name_lst={"potentials_sphinx_jth_default.csv"},
            )
            potential_df = potential_df[(potential_df["Model"] == "jth-gga-pbe")]
        else:
            raise ValueError(
                'The exchange correlation functional has to be set to "PBE" currently there are no "LDA" potentials.'
            )
        super(SphinxJTHPotentialFile, self).__init__(
            potential_df=potential_df,
            default_df=default_df,
            selected_atoms=selected_atoms,
        )

    def add_new_element(self, parent_element, new_element):
        """
        Adding a new user defined element with a different POTCAR file. It is assumed that the file exists

        Args:
            parent_element (str): Parent element
            new_element (str): Name of the new element (the name of the folder where the new POTCAR file exists

        """
        ds = self.find_default(element=parent_element)
        ds["Species"].values[0][0] = new_element
        path_list = ds["Filename"].values[0][0].split("/")
        path_list[-2] = new_element
        name_list = ds["Name"].values[0].split("-")
        name_list[0] = new_element
        ds["Name"].values[0] = "-".join(name_list)
        ds["Filename"].values[0][0] = "/".join(path_list)
        self._potential_df = self._potential_df.append(ds)
        ds = pandas.Series()
        ds.name = new_element
        ds["Name"] = "-".join(name_list)
        self._default_df = self._default_df.append(ds)



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import warnings

import numpy as np

from pyiron_atomistics.atomistics.job.atomistic import (
    AtomisticGenericJob,
)
from pyiron_atomistics.atomistics.job.atomistic import (
    MapFunctions as AtomisticMapFunctions,
)
from pyiron_atomistics.dft.waves.electronic import ElectronicStructure

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


class GenericDFTJob(AtomisticGenericJob):
    def __init__(self, project, job_name):
        super(GenericDFTJob, self).__init__(project, job_name)
        self._generic_input["fix_symmetry"] = True
        self.map_functions = MapFunctions()
        self._generic_input["k_mesh_spacing"] = None
        self._generic_input["k_mesh_center_shift"] = None
        self._generic_input["reduce_kpoint_symmetry"] = True

    @property
    def encut(self):
        return self.plane_wave_cutoff

    @encut.setter
    def encut(self, val):
        self.plane_wave_cutoff = val

    @property
    def kpoint_mesh(self):
        return self.get_kpoints()

    @kpoint_mesh.setter
    def kpoint_mesh(self, val):
        self.set_kpoints(mesh=val)

    @property
    def xc(self):
        return self.exchange_correlation_functional

    @xc.setter
    def xc(self, val):
        self.exchange_correlation_functional = val

    @property
    def plane_wave_cutoff(self):
        raise NotImplementedError(
            "The encut property is not implemented for this code."
        )

    @plane_wave_cutoff.setter
    def plane_wave_cutoff(self, val):
        raise NotImplementedError(
            "The encut property is not implemented for this code."
        )

    @property
    def spin_constraints(self):
        raise NotImplementedError(
            "The spin_constraints property is not implemented for this code."
        )

    @spin_constraints.setter
    def spin_constraints(self, val):
        raise NotImplementedError(
            "The spin_constraints property is not implemented for this code."
        )

    @property
    def exchange_correlation_functional(self):
        raise NotImplementedError(
            "The exchange property is not implemented for this code."
        )

    @exchange_correlation_functional.setter
    def exchange_correlation_functional(self, val):
        raise NotImplementedError(
            "The exchange property is not implemented for this code."
        )

    @property
    def k_mesh_spacing(self):
        """
        Number of unreduced k-points per Angstrom of the lattice vector

        Returns:
            float: Number of k-points per Angstrom
        """
        return self._generic_input["k_mesh_spacing"]

    @k_mesh_spacing.setter
    def k_mesh_spacing(self, val):
        self._generic_input["k_mesh_spacing"] = val

    @property
    def k_mesh_center_shift(self):
        """
        Number of unreduced k-points per Angstrom of the lattice vector

        Returns:
            float: Number of k-points per Angstrom
        """
        return self._generic_input["k_mesh_center_shift"]

    @k_mesh_center_shift.setter
    def k_mesh_center_shift(self, val):
        self._generic_input["k_mesh_center_shift"] = val

    @property
    def reduce_kpoint_symmetry(self):
        """
        Number of unreduced k-points per Angstrom of the lattice vector

        Returns:
            float: Number of k-points per Angstrom
        """
        return self._generic_input["reduce_kpoint_symmetry"]

    @reduce_kpoint_symmetry.setter
    def reduce_kpoint_symmetry(self, boolean):
        self._generic_input["reduce_kpoint_symmetry"] = boolean

    @property
    def fix_spin_constraint(self):
        return self._generic_input["fix_spin_constraint"]

    @fix_spin_constraint.setter
    def fix_spin_constraint(self, boolean):
        if not isinstance(boolean, bool):
            raise AssertionError()
        self._generic_input["fix_spin_constraint"] = boolean

    @property
    def fix_symmetry(self):
        return self._generic_input["fix_symmetry"]

    @fix_symmetry.setter
    def fix_symmetry(self, boolean):
        if not isinstance(boolean, bool):
            raise AssertionError()
        self._generic_input["fix_symmetry"] = boolean

    def _get_structure(self, frame=-1, wrap_atoms=True):
        snapshot = super(GenericDFTJob, self)._get_structure(
            frame=frame, wrap_atoms=wrap_atoms
        )
        spins = self.get_magnetic_moments(iteration_step=frame)
        if spins is not None:
            snapshot.set_initial_magnetic_moments(spins)
        return snapshot

    def set_mixing_parameters(
        self,
        method=None,
        n_pulay_steps=None,
        density_mixing_parameter=None,
        spin_mixing_parameter=None,
        density_residual_scaling=None,
        spin_residual_scaling=None,
    ):
        """
        Args:
            method (str): mixing method 'PULAY' or 'KERKER' (default: PULAY)
            n_pulay_steps (int): number of previous densities to use for the Pulay mixing
                (default: 7)
            density_mixing_parameter (float): mixing ratio of rho_opt to rho_in
            spin_mixing_parameter (float): linear mixing parameter for spin densities
            density_residual_scaling (float): scaling for the residual contribution of the Kerker
                mixing. 1 means it takes as much rho_in (input density) as residual (where Kerker
                preconditioning applies). The lower the value, the smaller the residual
                contribution becomes.
            spin_residual_scaling (float): scaling for the spin residual
                (cf. density_residual_scaling)

            Mixing ratio m is given by:

                rho^n = (m-1)*rho_in+m*preconditioner*rho_opt


        A low value of density mixing parameter may lead to a more stable convergence, but will slow
        down the calculation if set too low. In systems with bands around the Fermi energy (i.e.
        metals, e.g. Fe, Mn), the mixing parameters should be small. In insulators (e.g. Al, Si) the
        mixing parameter should be high.

        """
        raise NotImplementedError(
            "set_mixing_parameters is not implemented for this code."
        )

    def restart_for_band_structure_calculations(self, job_name=None):
        """
        Restart a new job created from an existing calculation by reading the charge density
        for band structure calculations.

        Args:
            job_name (str/None): Job name

        Returns:
            new_ham (pyiron_atomistics.dft.job.generic.GenericDFTJob): New job
        """
        raise NotImplementedError(
            "restart_for_band_structure_calculations is not implemented for this code."
        )

    def get_magnetic_moments(self, iteration_step=-1):
        """
        Gives the magnetic moments of a calculation for each iteration step.

        Args:
            iteration_step (int): Step for which the structure is requested

        Returns:
            numpy.ndarray/None: array of final magmetic moments or None if no magnetic moment is given
        """
        spins = self.get("output/generic/dft/atom_spins")
        if spins is not None:
            return spins[iteration_step]
        else:
            return None

    def get_kpoints(self):
        raise NotImplementedError(
            "The get_kpoints() function is not implemented for this code."
        )

    def get_valence_and_total_charge_density(self):
        raise NotImplementedError(
            "The get_valence_and_total_charge_density() function is not implemented for this code."
        )

    def get_k_mesh_by_cell(self, k_mesh_spacing, cell=None):
        """
        Get k-mesh density according to the box size.

        Args:
            k_mesh_spacing (float): K-point spacing in units of 2 * pi reciprocal Angstrom.
                                (smaller values result in a denser mesh for a given structure).
            cell (numpy.ndarray/list): The cell shape

        Returns:
            list/numpy.ndarray: Mesh size

        """
        if cell is None:
            if self.structure is None:
                raise ValueError(
                    "Can't generate k-points without structure being set and if cell is not specified"
                )
            cell = self.structure.cell
        return get_k_mesh_by_density(cell=cell, k_mesh_spacing=k_mesh_spacing)

    def set_kpoints(
        self,
        mesh=None,
        scheme="MP",
        center_shift=None,
        symmetry_reduction=True,
        manual_kpoints=None,
        weights=None,
        reciprocal=True,
        k_mesh_spacing=None,
        n_path=None,
        path_name=None,
    ):
        """
        Function to setup the k-points

        Args:
            mesh (list/numpy.ndarray): Size of the mesh (ignored if scheme is not set to 'MP' or kpoints_per_reciprocal_
            angstrom is set)
            scheme (str): Type of k-point generation scheme (MP/GP(gamma point)/Manual/Line)
            center_shift (list/numpy.ndarray/None): Shifts the center of the mesh from the gamma point by the given vector in relative coordinates
            symmetry_reduction (boolean): Tells if the symmetry reduction is to be applied to the k-points
            manual_kpoints (list/numpy.ndarray): Manual list of k-points
            weights(list/numpy.ndarray): Manually supplied weights to each k-point in case of the manual mode
            reciprocal (bool): Tells if the supplied values are in reciprocal (direct) or cartesian coordinates (in
            reciprocal space)
            k_mesh_spacing (float): K-point spacing in units of 2 * pi reciprocal Angstrom.
                                (smaller values result in a denser mesh for a given structure).
            n_path (int): Number of points per trace part for line mode
            path_name (str): Name of high symmetry path used for band structure calculations.
        """
        if k_mesh_spacing is not None:
            if mesh is not None:
                warnings.warn("mesh value is overwritten by k_mesh_spacing")
            mesh = self.get_k_mesh_by_cell(k_mesh_spacing=k_mesh_spacing)
        self.k_mesh_spacing = k_mesh_spacing
        self.k_mesh_center_shift = center_shift
        self.reduce_kpoint_symmetry = symmetry_reduction
        if mesh is not None:
            if np.min(mesh) <= 0:
                raise ValueError("mesh values must be larger than 0")
        if center_shift is not None:
            if np.min(center_shift) < 0 or np.max(center_shift) > 1:
                warnings.warn("center_shift is given in relative coordinates")
        self._set_kpoints(
            mesh=mesh,
            scheme=scheme,
            center_shift=center_shift,
            symmetry_reduction=symmetry_reduction,
            manual_kpoints=manual_kpoints,
            weights=weights,
            reciprocal=reciprocal,
            n_path=n_path,
            path_name=path_name,
        )

    def calc_static(
        self,
        electronic_steps=60,
    ):
        super().calc_static()

    def calc_minimize(
        self,
        electronic_steps=60,
        ionic_steps=100,
        max_iter=None,
        pressure=None,
        ionic_energy_tolerance=None,
        ionic_force_tolerance=None,
        ionic_energy=None,
        ionic_forces=None,
        volume_only=False,
    ):
        super().calc_minimize(max_iter=max_iter, pressure=pressure)

    def calc_md(
        self,
        temperature=None,
        n_ionic_steps=1000,
        n_print=1,
        time_step=1.0,
        **kwargs,
    ):
        self._generic_input["fix_symmetry"] = False
        super().calc_md(
            temperature=temperature,
            n_ionic_steps=n_ionic_steps,
            n_print=n_print,
            time_step=time_step,
        )

    def nbands_convergence_check(self):
        """
        Function to check if there are a sufficient number of empty bands in the calculation to ensure electronic convergence.

        Returns:

            bool : True if the highest band is unoccupied, False if the highest band is occupied
        """
        return np.all(
            np.isclose(self["output/electronic_structure/occ_matrix"][:, :, -1], 0)
        )  # shape is n_spin x n_kpoints x n_bands

    # Backward compatibility
    def get_encut(self):
        return self.encut

    def set_encut(self, encut):
        """
        Sets the plane-wave energy cutoff
        Args:
            encut (float): The energy cutoff in eV
        """
        self.plane_wave_cutoff = encut

    def set_exchange_correlation_functional(self, exchange_correlation_functional):
        self.exchange_correlation_functional = exchange_correlation_functional

    def set_empty_states(self, n_empty_states=None):
        raise NotImplementedError(
            "The set_empty_states function is not implemented for this code."
        )

    def get_bader_charges(self):
        """
        Returns the total charge on every atom determined by the Bader charge partitioning scheme.

        Returns:
            numpy.ndarray: The Bader charges for each atom

        """
        return self["output/generic/dft/bader_charges"]

    def get_bader_volumes(self):
        """
        Returns the integration volume around every atom from the Bader charge partitioning scheme.

        Returns:
            numpy.ndarray: The Bader charges for each atom

        """
        return self["output/generic/dft/bader_volumes"]

    def _set_kpoints(
        self,
        mesh=None,
        scheme="MP",
        center_shift=None,
        symmetry_reduction=True,
        manual_kpoints=None,
        weights=None,
        reciprocal=True,
        n_path=None,
        path_name=None,
    ):
        raise NotImplementedError(
            "The set_kpoints function is not implemented for this code."
        )

    def get_electronic_structure(self):
        """
        Gets the electronic structure instance from the hdf5 file

        Returns:
                pyiron_atomistics.atomistics.waves.electronic.ElectronicStructure instance
        """
        if self.status not in ["finished", "warning", "not_converged"]:
            return
        else:
            with self.project_hdf5.open("output") as ho:
                es_obj = ElectronicStructure()
                es_obj.from_hdf(ho)
            return es_obj

    def modify_kpoints(self):
        if self.k_mesh_spacing is not None:
            self.set_kpoints(
                center_shift=self.k_mesh_center_shift,
                k_mesh_spacing=self.k_mesh_spacing,
                symmetry_reduction=self.reduce_kpoint_symmetry,
            )

    def save(self):
        self.modify_kpoints()
        super(GenericDFTJob, self).save()

    def get_density_of_states(self, sigma=0.1, shift_by_fermi_energy=True, grid=None):
        """
        Get density of states from a fully converged result. A Gaussian smeared histogram is
        returned

        Args:
            sigma (float): Gaussian smearing parameter in energy unit.
            shift_by_fermi_energy (bool): Shift the histogram by the Fermi energy value. Setting
                this to False will return code specific absolute values which have physically no
                meaning.
            grid (list/numpy.ndarray): Energy grid. If None, the interval between maximum and
                minimum eigenvalues plus 5 x sigma with a step length of sigma will be taken.

        Returns:
            (dict): grid and density of states (n_spin x energy_grid)
        """
        if sigma <= 0:
            raise ValueError("Sigma must be a positive float")
        k_weights = self["output/electronic_structure/k_weights"]
        if k_weights is None:
            raise ValueError("k-point weighting not found")
        e_fermi = 0
        if shift_by_fermi_energy:
            e_fermi = self["output/electronic_structure/efermi"]
        eigen_values = self["output/electronic_structure/eig_matrix"]
        eigen_values = eigen_values - e_fermi
        if grid is None:
            grid = np.arange(
                eigen_values.min() - 5 * sigma, eigen_values.max() + 5 * sigma, sigma
            )
        hist = (
            eigen_values[:, :, :, np.newaxis]
            - grid[np.newaxis, np.newaxis, np.newaxis, :]
        )
        hist = (
            np.exp(-((hist) ** 2) / (2 * sigma**2))
            * k_weights[np.newaxis, :, np.newaxis, np.newaxis]
        )
        hist = np.sum(hist, axis=(1, 2))
        hist *= 2 / len(eigen_values) / np.sqrt(2 * np.pi * sigma**2)
        return {"grid": grid, "dos": hist}


def get_k_mesh_by_density(cell, k_mesh_spacing=0.5):
    """
    Get k-mesh density according to the box size.

    Args:
        cell (numpy.ndarray/list): The cell shape
        k_mesh_spacing (float): K-point spacing in units of 2 * pi reciprocal Angstrom.
                                (smaller values result in a denser mesh for a given structure).

    Returns:
        list/numpy.ndarray: Mesh size

    """
    omega = np.linalg.det(cell)
    l1, l2, l3 = cell
    g1 = 2 * np.pi / omega * np.cross(l2, l3)
    g2 = 2 * np.pi / omega * np.cross(l3, l1)
    g3 = 2 * np.pi / omega * np.cross(l1, l2)

    kmesh = np.rint(
        np.array([np.linalg.norm(g) for g in [g1, g2, g3]]) / k_mesh_spacing
    )
    kmesh[kmesh < 1] = 1
    return [int(k) for k in kmesh]


def set_encut(job, parameter):
    job.set_encut(parameter)
    return job


def set_kpoints(job, parameter):
    job.set_kpoints(parameter)
    return job


class MapFunctions(AtomisticMapFunctions):
    def __init__(self):
        super().__init__()
        self.set_encut = set_encut
        self.set_kpoints = set_kpoints



import copy
import os
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
import yaml
from ase.io import read
from pyiron_base import DataContainer, GenericJob
from pyiron_snippets.import_alarm import ImportAlarm

from pyiron_atomistics.atomistics.structure.atoms import Atoms, ase_to_pyiron
from pyiron_atomistics.atomistics.structure.has_structure import HasStructure
from pyiron_atomistics.lammps.potential import LammpsPotential, LammpsPotentialFile
from pyiron_atomistics.lammps.structure import (
    LammpsStructure,
    UnfoldingPrism,
    structure_to_lammps,
)

calphy_version = "1.0.0"

with ImportAlarm(
    "Calphy functionality requires the `calphy` module (and its dependencies) specified as extra"
    "requirements. Please install it and try again."
) as calphy_alarm:
    from calphy import Alchemy, Calculation, Liquid, Solid
    from calphy import __version__ as calphy_version
    from calphy.routines import routine_alchemy, routine_fe, routine_pscale, routine_ts
    from pyscal.trajectory import Trajectory as PyscalTrajectory

__author__ = "Sarath Menon"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sarath Menon"
__email__ = "s.menon@mpie.de"
__status__ = "development"
__date__ = "April 1, 2022"


class Calphy(GenericJob, HasStructure):
    """
    Class to set up and run calphy jobs for calculation of free energies using LAMMPS.

    An input structure (:attr:`structure`) and interatomic potential (:attr:`potential`) are necessary input options. The additional input options such as the temperature and pressure are specified in the :meth:`.calc_free_energy` method. Depending on the input parameters, a corresponding calculation mode is selected. Further input options can be accessed through :attr:`input.md` and :attr:`input.tolerance`.

    An example which calculates the free energy of Cu using an interatomic potential:

    ```python
    job.structure = pr.create.structure.ase.bulk('Cu', cubic=True).repeat(5)
    job.potential = "2001--Mishin-Y--Cu-1--LAMMPS--ipr1"
    job.calc_free_energy(temperature=1100, pressure=0, reference_phase="solid")
    job.run()
    ```

    In order to calculate the free energy of the liquid phase, the `reference_phase` should be set to `liquid`.

    The different modes can be selected as follows:

    For free energy at a given temperature and pressure:

    ```python
    job.calc_free_energy(temperature=1100, pressure=0, reference_phase="solid")
    ```

    Alternatively, :func:`calc_mode_fe` can be used.

    To obtain the free energy between a given temperature range (temperature scaling):

    ```python
    job.calc_free_energy(temperature=[1100, 1400], pressure=0, reference_phase="solid")
    ```

    Alternatively, :func:`calc_mode_ts` can be used.

    For free energy between a given pressure range (pressure scaling)

    ```python
    job.calc_free_energy(temperature=1000, pressure=[0, 100000], reference_phase="solid")
    ```

    Alternatively, :func:`calc_mode_pscale` can be used.

    To obtain the free energy difference between two interatomic potentials (alchemy/upsampling)

    ```python
    job.potential = ["2001--Mishin-Y--Cu-1--LAMMPS--ipr1", "1986--Foiles-S-M--Cu--LAMMPS--ipr1"]
    job.calc_free_energy(temperature=1100, pressure=0, reference_phase="solid")
    job.run()
    ```

    Alternatively, :func:`calc_mode_alchemy` can be used.

    The way `pressure` is specified determines how the barostat affects the system. For isotropic pressure control:

    ```python
    job.calc_free_energy(temperature=[1100, 1400], pressure=0, reference_phase="solid")
    ```

    For anisotropic pressure control:

    ```python
    job.calc_free_energy(temperature=[1100, 1400], pressure=[0, 0, 0], reference_phase="solid")
    ```

    To constrain the lattice:

    ```python
    job.calc_free_energy(temperature=[1100, 1400], pressure=None, reference_phase="solid")
    ```

    In addition the boolean option :attr:`input.npt` can be used to determine the MD ensemble. If True, temperature integration and alchemy/upsampling are carried out in the NPT ensemble. If False, the NVT ensemble is employed.

    After the calculation is over, the various output options can be accessed through `job.output`.

    Specialised output depending on the selected mode is also available. For example the energy difference between the system of
    interest and the reference system for mode `fe` (free energy calculation), is available under `job.output.fe`. Similarly other output
    corresponding to modes such as temperature scaling and pressure scaling can be found under `job.output.ts` and `job.output.ps`.
    """

    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        self.input = DataContainer(self._default_input, table_name="inputdata")
        self._potential_initial = None
        self._potential_final = None
        self.input.potential_initial_name = None
        self.input.potential_final_name = None
        self.input.structure = None
        self._output = DataContainer(table_name="output")
        self.input._pot_dict_initial = None
        self.input._pot_dict_final = None
        self.__version__ = calphy_version

    @property
    def output(self):
        # LHS - status from hdf5, RHS - status from memory
        if (self["status"] == "finished") and (
            self.status._get_status_from_dict() != "finished"
        ):
            self.from_hdf()
            self.status.finished = True
        return self._output

    @property
    def _default_input(self):
        return {
            "mode": None,
            "pressure": 0,
            "temperature": 0,
            "reference_phase": None,
            "npt": None,
            "n_equilibration_steps": 15000,
            "n_switching_steps": 25000,
            "n_print_steps": 1000,
            "n_iterations": 1,
            "spring_constants": None,
            "equilibration_control": "nose-hoover",
            "melting_cycle": True,
            "md": {
                "timestep": 0.001,
                "n_small_steps": 10000,
                "n_every_steps": 10,
                "n_repeat_steps": 10,
                "n_cycles": 100,
                "thermostat_damping": 0.5,
                "barostat_damping": 0.1,
            },
            "tolerance": {
                "lattice_constant": 0.0002,
                "spring_constant": 0.01,
                "solid_fraction": 0.7,
                "liquid_fraction": 0.05,
                "pressure": 0.5,
            },
            "nose_hoover": {
                "thermostat_damping": 0.1,
                "barostat_damping": 0.1,
            },
            "berendsen": {
                "thermostat_damping": 100.0,
                "barostat_damping": 100.0,
            },
        }

    def set_potentials(self, potential_filenames: Union[list, str]):
        """
        Set the interatomic potential from a given name

        Args:
            potential_filenames (list, str): list of filenames

        Returns:
            None
        """
        if not isinstance(potential_filenames, list):
            potential_filenames = [potential_filenames]
        if len(potential_filenames) > 0:
            if isinstance(potential_filenames[0], pd.DataFrame):
                potential = potential_filenames[0]
                self.input._pot_dict_initial = potential  # .to_dict()
            else:
                potential = LammpsPotentialFile().find_by_name(potential_filenames[0])
                self.input.potential_initial_name = potential_filenames[0]
            self._potential_initial = LammpsPotential()
            self._potential_initial.df = potential

        if len(potential_filenames) > 1:
            if isinstance(potential_filenames[1], pd.DataFrame):
                potential = potential_filenames[1]
                self.input._pot_dict_final = potential  # .to_dict()
            else:
                potential = LammpsPotentialFile().find_by_name(potential_filenames[1])
                self.input.potential_final_name = potential_filenames[1]
            self._potential_final = LammpsPotential()
            self._potential_final.df = potential
        if len(potential_filenames) > 2:
            raise ValueError("Maximum two potentials can be provided")

    def get_potentials(self) -> List[str]:
        """
        Return the interatomic potentials

        Args:
            None

        Returns:
            list of str: list of interatomic potentials
        """
        if self._potential_final is None:
            return [self._potential_initial.df]
        else:
            return [self._potential_initial.df, self._potential_final.df]

    def _copy_pot_files(self):
        """
        Copy potential files to the working directory

        Args:
            None

        Returns:
            None
        """
        if self._potential_initial is not None:
            self._potential_initial.copy_pot_files(self.working_directory)
        if self._potential_final is not None:
            self._potential_final.copy_pot_files(self.working_directory)

    def _prepare_pair_styles(self) -> Tuple[List, List]:
        """
        Prepare pair style and pair coeff

        Args:
            None

        Returns:
            list: pair style and pair coeff
        """

        pair_style = []
        pair_coeff = []

        if self._potential_initial is not None:
            pair_style.append(
                " ".join(
                    self._potential_initial.df["Config"]
                    .to_list()[0][0]
                    .strip()
                    .split()[1:]
                )
            )
            pair_coeff.append(
                " ".join(
                    self._potential_initial.df["Config"]
                    .to_list()[0][1]
                    .strip()
                    .split()[1:]
                )
            )
        if self._potential_final is not None:
            pair_style.append(
                " ".join(
                    self._potential_final.df["Config"]
                    .to_list()[0][0]
                    .strip()
                    .split()[1:]
                )
            )
            pair_coeff.append(
                " ".join(
                    self._potential_final.df["Config"]
                    .to_list()[0][1]
                    .strip()
                    .split()[1:]
                )
            )

        return pair_style, pair_coeff

    def _get_element_list(self) -> List[str]:
        """
        Get elements as defined in pair style

        Args:
            None

        Returns:
            list: symbols of the elements
        """
        elements = self._potential_initial.get_element_lst()
        return elements

    def _get_masses(self) -> List[float]:
        """
        Get masses as defined in pair style

        Args:
            None

        Returns:
            list: masses of the elements
        """
        elements_from_pot = self._potential_initial.get_element_lst()
        elements_object_lst = self.structure.get_species_objects()
        elements_struct_lst = self.structure.get_species_symbols()

        masses = []
        for element_name in elements_from_pot:
            if element_name in elements_struct_lst:
                index = list(elements_struct_lst).index(element_name)
                masses.append(elements_object_lst[index].AtomicMass)
            else:
                masses.append(1.0)
        return masses

    def _potential_from_hdf(self):
        """
        Recreate the potential from filename stored in hdf5

        Args:
            None

        Returns:
            None
        """
        filenames = []
        if self.input.potential_initial_name is not None:
            filenames.append(self.input.potential_initial_name)
        elif self.input._pot_dict_initial is not None:
            filenames.append(pd.DataFrame(data=self.input._pot_dict_initial))
        if self.input.potential_final_name is not None:
            filenames.append(self.input.potential_final_name)
        elif self.input._pot_dict_final is not None:
            filenames.append(pd.DataFrame(data=self.input._pot_dict_final))

        self.set_potentials(filenames)

    @property
    def potential(self):
        potentials = self.get_potentials()
        if len(potentials) == 1:
            return potentials[0]
        return potentials

    @potential.setter
    def potential(self, potential_filenames):
        self.set_potentials(potential_filenames)

    @property
    def structure(self):
        return self.input.structure

    @structure.setter
    def structure(self, val):
        self.input.structure = val

    def view_potentials(self) -> List:
        """
        View a list of available interatomic potentials

        Args:
            None

        Returns:
            list: list of available potentials
        """
        if not self.structure:
            raise ValueError("please assign a structure first")
        else:
            list_of_elements = set(self.structure.get_chemical_symbols())
        list_of_potentials = LammpsPotentialFile().find(list_of_elements)
        if list_of_potentials is not None:
            return list_of_potentials
        else:
            raise TypeError(
                "No potentials found for this kind of structure: ",
                str(list_of_elements),
            )

    def list_potentials(self):
        """
        List of interatomic potentials suitable for the current atomic structure.

        use self.potentials_view() to get more details.

        Args:
            None

        Returns:
            list: potential names
        """
        return list(self.view_potentials()["Name"].values)

    def write_structure(self, structure, file_name: str, working_directory: str):
        """
        Write structure to file

        Args:
            structure: input structure
            file_name (str): output file name
            working_directory (str): output working directory

        Returns:
            None
        """
        lmp_structure = LammpsStructure()
        lmp_structure.potential = self._potential_initial
        lmp_structure.atom_type = "atomic"
        lmp_structure.el_eam_lst = self._potential_initial.get_element_lst()
        lmp_structure.structure = structure_to_lammps(structure)

        if not set(lmp_structure.structure.get_species_symbols()).issubset(
            set(lmp_structure.el_eam_lst)
        ):
            raise ValueError(
                "The selected potentials do not support the given combination of elements."
            )
        lmp_structure.write_file(file_name=file_name, cwd=working_directory)

    def _determine_mode(self):
        """
        Determine the calculation mode

        Args:
            None

        Returns:
            None
        """
        if len(self.get_potentials()) == 2:
            self.input.mode = "alchemy"
            self.input.reference_phase = "alchemy"
        elif isinstance(self.input.pressure, list):
            if len(self.input.pressure) == 2:
                self.input.mode = "pscale"
        elif isinstance(self.input.temperature, list):
            if len(self.input.temperature) == 2:
                self.input.mode = "ts"
        else:
            self.input.mode = "fe"
        # if mode was not set, raise Error
        if self.input.mode is None:
            raise RuntimeError("Could not determine the mode")

    def _create_calc(self):
        """
        Create a calc object
        """
        calc = copy.deepcopy(self._default_input)

        for key in self._default_input.keys():
            if key not in ["md", "tolerance", "nose_hoover", "berendsen"]:
                calc[key] = self.input[key]

        for key in self._default_input["md"].keys():
            calc["md"][key] = self.input["md"][key]

        for key in self._default_input["tolerance"].keys():
            calc["tolerance"][key] = self.input["tolerance"][key]

        for key in self._default_input["nose_hoover"].keys():
            calc["nose_hoover"][key] = self.input["nose_hoover"][key]

        for key in self._default_input["berendsen"].keys():
            calc["berendsen"][key] = self.input["berendsen"][key]

        calc["lattice"] = os.path.join(self.working_directory, "conf.data")

        pair_style, pair_coeff = self._prepare_pair_styles()
        calc["fix_potential_path"] = False
        calc["pair_style"] = pair_style
        calc["pair_coeff"] = pair_coeff

        calc["element"] = self._get_element_list()
        calc["mass"] = self._get_masses()

        calc["queue"] = {}
        calc["queue"]["cores"] = self.server.cores

        calculations = {}
        calculations["calculations"] = [calc]

        with open(os.path.join(self.working_directory, "input.yaml"), "w") as fout:
            yaml.safe_dump(calculations, fout)

    def write_input(self):
        """
        Write input for calphy calculation

        Args:
            None

        Returns:
            None
        """
        file_name = "conf.data"
        self.write_structure(self.structure, file_name, self.working_directory)
        self._copy_pot_files()
        self._create_calc()

    def calc_mode_fe(
        self,
        temperature: float = None,
        pressure: Union[list, float, None] = None,
        reference_phase: str = None,
        n_equilibration_steps: int = 15000,
        n_switching_steps: int = 25000,
        n_print_steps: int = 0,
        n_iterations: int = 1,
    ):
        """
        Calculate free energy at given conditions

        Args:
            None

        Returns:
            None
        """
        if temperature is None:
            raise ValueError("provide a temperature")
        if reference_phase is None:
            raise ValueError("provide a reference_phase")

        self.input.temperature = temperature
        self.input.pressure = pressure
        self.input.npt = pressure is not None
        self.input.reference_phase = reference_phase
        self.input.n_equilibration_steps = n_equilibration_steps
        self.input.n_switching_steps = n_switching_steps
        self.input.n_print_steps = n_print_steps
        self.input.n_iterations = n_iterations
        self.input.mode = "fe"

    def calc_mode_ts(
        self,
        temperature: float = None,
        pressure: Union[list, float, None] = None,
        reference_phase: str = None,
        n_equilibration_steps: int = 15000,
        n_switching_steps: int = 25000,
        n_print_steps: int = 0,
        n_iterations: int = 1,
    ):
        """
        Calculate free energy between given temperatures

        Args:
            None

        Returns:
            None
        """
        if temperature is None:
            raise ValueError("provide a temperature")
        if reference_phase is None:
            raise ValueError("provide a reference_phase")

        self.input.temperature = temperature
        self.input.pressure = pressure
        self.input.npt = pressure is not None
        self.input.reference_phase = reference_phase
        self.input.n_equilibration_steps = n_equilibration_steps
        self.input.n_switching_steps = n_switching_steps
        self.input.n_print_steps = n_print_steps
        self.input.n_iterations = n_iterations
        self.input.mode = "ts"

    def calc_mode_alchemy(
        self,
        temperature: float = None,
        pressure: Union[list, float, None] = None,
        reference_phase: str = None,
        n_equilibration_steps: int = 15000,
        n_switching_steps: int = 25000,
        n_print_steps: int = 0,
        n_iterations: int = 1,
    ):
        """
        Perform upsampling/alchemy between two interatomic potentials

        Args:
            None

        Returns:
            None
        """
        if temperature is None:
            raise ValueError("provide a temperature")
        self.input.temperature = temperature
        self.input.pressure = pressure
        self.input.npt = pressure is not None
        self.input.reference_phase = reference_phase
        self.input.n_equilibration_steps = n_equilibration_steps
        self.input.n_switching_steps = n_switching_steps
        self.input.n_print_steps = n_print_steps
        self.input.n_iterations = n_iterations
        self.input.mode = "alchemy"

    def calc_mode_pscale(
        self,
        temperature: float = None,
        pressure: Union[list, float, None] = None,
        reference_phase: str = None,
        n_equilibration_steps: int = 15000,
        n_switching_steps: int = 25000,
        n_print_steps: int = 0,
        n_iterations: int = 1,
    ):
        """
        Calculate free energy between two given pressures

        Args:
            None

        Returns:
            None
        """
        if temperature is None:
            raise ValueError("provide a temperature")
        if reference_phase is None:
            raise ValueError("provide a reference_phase")

        self.input.temperature = temperature
        self.input.pressure = pressure
        self.input.npt = True
        self.input.reference_phase = reference_phase
        self.input.n_equilibration_steps = n_equilibration_steps
        self.input.n_switching_steps = n_switching_steps
        self.input.n_print_steps = n_print_steps
        self.input.n_iterations = n_iterations
        self.input.mode = "pscale"

    def calc_free_energy(
        self,
        temperature: float = None,
        pressure: Union[list, float, None] = None,
        reference_phase: str = None,
        n_equilibration_steps: int = 15000,
        n_switching_steps: int = 25000,
        n_print_steps: int = 0,
        n_iterations: int = 1,
    ):
        """
        Calculate free energy at given conditions

        Args:
            None

        Returns:
            None
        """
        if temperature is None:
            raise ValueError("provide a temperature")
        self.input.temperature = temperature
        self.input.pressure = pressure
        self.input.npt = pressure is not None
        self.input.reference_phase = reference_phase
        self.input.n_equilibration_steps = n_equilibration_steps
        self.input.n_switching_steps = n_switching_steps
        self.input.n_print_steps = n_print_steps
        self.input.n_iterations = n_iterations
        self._determine_mode()
        if self.input.mode != "alchemy":
            if reference_phase is None:
                raise ValueError("provide a reference_phase")

    def run_static(self):
        with open(os.path.join(self.working_directory, "input.yaml"), "r") as fin:
            calc = yaml.safe_load(fin)["calculations"][0]
        calc = Calculation(**calc)
        self.status.running = True
        if self.input.reference_phase == "alchemy":
            job = Alchemy(calculation=calc, simfolder=self.working_directory)
        elif self.input.reference_phase == "solid":
            job = Solid(calculation=calc, simfolder=self.working_directory)
        elif self.input.reference_phase == "liquid":
            job = Liquid(calculation=calc, simfolder=self.working_directory)
        else:
            raise ValueError("Unknown reference state")

        if self.input.mode == "alchemy":
            routine_alchemy(job)
        elif self.input.mode == "fe":
            routine_fe(job)
        elif self.input.mode == "ts":
            routine_ts(job)
        elif self.input.mode == "pscale":
            routine_pscale(job)
        else:
            raise ValueError("Unknown mode")
        self.run()
        self.status.collect = True
        self.run()

    def _get_structure(self, frame=-1, wrap_atoms=False):
        """ """
        return [self.structure, self.output.structure_final][frame]

    def _number_of_structures(self):
        return 2

    # @property
    # def output(self):
    #    if len(self._output) == 0:
    #        self.collect_output()
    #    return self._output

    def collect_general_output(self):
        """
        Collect the output from calphy

        Args:
            None

        Returns:
            None
        """
        reportfile = os.path.join(self.working_directory, "report.yaml")

        if os.path.exists(reportfile):
            with open(reportfile, "r") as fin:
                data = yaml.safe_load(fin)

            concentration = data["input"]["concentration"].split()
            concentration = [float(x) for x in concentration]

            if "spring_constant" in data["average"].keys():
                self._output["spring_constant"] = data["average"]["spring_constant"]
            if "density" in data["average"].keys():
                self._output["atomic_density"] = data["average"]["density"]
            self._output["atomic_volume"] = data["average"]["vol_atom"]

            # main results from mode fe
            self._output["temperature"] = self.input.temperature
            self._output["pressure"] = self.input.pressure
            self._output["energy_free"] = data["results"]["free_energy"]
            self._output["energy_free_error"] = data["results"]["error"]
            self._output["energy_free_harmonic_reference"] = data["results"][
                "reference_system"
            ]
            self._output["energy_work"] = data["results"]["work"]
            self._output["energy_pressure"] = data["results"]["pv"]

            # collect ediffs and so on
            f_ediff, b_ediff, flambda, blambda = self._collect_ediff(concentration)
            self._output["fe/forward/energy_diff"] = list(f_ediff)
            self._output["fe/backward/energy_diff"] = list(b_ediff)
            self._output["fe/forward/lambda"] = list(flambda)
            self._output["fe/backward/lambda"] = list(blambda)

            # get final structure
            pyiron_atoms = read(
                os.path.join(self.working_directory, "conf.equilibration.data"),
                format="lammps-data",
                style="atomic",
            )

            self.output["structure_final"] = ase_to_pyiron(pyiron_atoms)

            if self.input.mode == "ts":
                datfile = os.path.join(self.working_directory, "temperature_sweep.dat")
                t, fe, ferr = np.loadtxt(datfile, unpack=True, usecols=(0, 1, 2))

                # replace the quantities with updates ones
                self._output["energy_free"] = np.array(fe)
                self._output["energy_free_error"] = np.array(ferr)
                self._output["temperature"] = np.array(t)

                # collect diffs
                (
                    f_ediff,
                    b_ediff,
                    f_vol,
                    b_vol,
                    f_press,
                    b_press,
                    flambda,
                    blambda,
                ) = self._collect_thermo(mode="ts")
                self._output["ts/forward/energy_diff"] = list(f_ediff)
                self._output["ts/backward/energy_diff"] = list(b_ediff)
                self._output["ts/forward/lambda"] = list(flambda)
                self._output["ts/backward/lambda"] = list(blambda)
                self._output["ts/forward/volume"] = list(f_vol)
                self._output["ts/backward/volume"] = list(b_vol)
                self._output["ts/forward/pressure"] = list(f_press)
                self._output["ts/backward/pressure"] = list(b_press)

                # populate structures
                (
                    fwd_positions,
                    bkd_positions,
                    fwd_cells,
                    bkd_cells,
                ) = self._get_positions()
                self._output["ts/forward/positions"] = fwd_positions
                self._output["ts/backward/positions"] = bkd_positions
                self._output["ts/forward/cells"] = fwd_cells
                self._output["ts/backward/cells"] = bkd_cells

            elif self.input.mode == "pscale":
                datfile = os.path.join(self.working_directory, "pressure_sweep.dat")
                p, fe, ferr = np.loadtxt(datfile, unpack=True, usecols=(0, 1, 2))
                self._output["energy_free"] = np.array(fe)
                self._output["energy_free_error"] = np.array(ferr)
                self._output["pressure"] = np.array(p)

                (
                    f_ediff,
                    b_ediff,
                    f_vol,
                    b_vol,
                    f_press,
                    b_press,
                    flambda,
                    blambda,
                ) = self._collect_thermo(mode="pscale")
                self._output["ps/forward/energy_diff"] = list(f_ediff)
                self._output["ps/backward/energy_diff"] = list(b_ediff)
                self._output["ps/forward/lambda"] = list(flambda)
                self._output["ps/backward/lambda"] = list(blambda)
                self._output["ps/forward/volume"] = list(f_vol)
                self._output["ps/backward/volume"] = list(b_vol)
                self._output["ps/forward/pressure"] = list(f_press)
                self._output["ps/backward/pressure"] = list(b_press)

    def _collect_ediff(self, concentration):
        """
        Calculate the energy difference between reference system and system of interest

        Args:
            None

        Returns:
            None
        """
        f_ediff = []
        b_ediff = []

        for i in range(1, self.input.n_iterations + 1):
            fwdfilename = os.path.join(self.working_directory, "forward_%d.dat" % i)
            bkdfilename = os.path.join(self.working_directory, "backward_%d.dat" % i)
            nelements = len(self._get_element_list())

            if self.input.reference_phase == "solid":
                fdui = np.loadtxt(fwdfilename, unpack=True, comments="#", usecols=(0,))
                bdui = np.loadtxt(bkdfilename, unpack=True, comments="#", usecols=(0,))

                fdur = np.zeros(len(fdui))
                bdur = np.zeros(len(bdui))

                for i in range(nelements):
                    fdur += concentration[i] * np.loadtxt(
                        fwdfilename, unpack=True, comments="#", usecols=(i + 1,)
                    )
                    bdur += concentration[i] * np.loadtxt(
                        bkdfilename, unpack=True, comments="#", usecols=(i + 1,)
                    )

                flambda = np.loadtxt(
                    fwdfilename, unpack=True, comments="#", usecols=(nelements + 1,)
                )
                blambda = np.loadtxt(
                    bkdfilename, unpack=True, comments="#", usecols=(nelements + 1,)
                )
            else:
                fdui, fdur, flambda = np.loadtxt(
                    fwdfilename, unpack=True, comments="#", usecols=(0, 1, 2)
                )
                bdui, bdur, blambda = np.loadtxt(
                    bkdfilename, unpack=True, comments="#", usecols=(0, 1, 2)
                )

            f_ediff.append(fdui - fdur)
            b_ediff.append(bdui - bdur)

        return f_ediff, b_ediff, flambda, blambda

    def _collect_thermo(self, mode="ts"):
        """
        Collect thermo quantities after ts run
        """
        f_ediff = []
        b_ediff = []
        f_vol = []
        b_vol = []
        f_press = []
        b_press = []

        for i in range(1, self.input.n_iterations + 1):
            fwdfilename = os.path.join(
                self.working_directory, f"{mode}.forward_{i}.dat"
            )
            bkdfilename = os.path.join(
                self.working_directory, f"{mode}.backward_{i}.dat"
            )

            fdx, fp, fvol, flambda = np.loadtxt(fwdfilename, unpack=True, comments="#")
            bdx, bp, bvol, blambda = np.loadtxt(bkdfilename, unpack=True, comments="#")

            fdx /= flambda
            bdx /= blambda

            f_ediff.append(fdx)
            b_ediff.append(bdx)

            f_vol.append(fvol)
            b_vol.append(bvol)

            f_press.append(fp)
            b_press.append(bp)

        return f_ediff, b_ediff, f_vol, b_vol, f_press, b_press, flambda, blambda

    def _get_positions(self):
        """
        Collect positions and cells
        """
        fwd_positions = []
        bkd_positions = []
        fwd_cells = []
        bkd_cells = []

        for i in range(1, self.input.n_iterations + 1):
            fwdfilename = os.path.join(
                self.working_directory, f"traj.ts.forward_{i}.dat"
            )
            bkdfilename = os.path.join(
                self.working_directory, f"traj.ts.backward_{i}.dat"
            )

            fp = []
            fc = []
            bp = []
            bc = []

            if os.path.exists(fwdfilename):
                traj = PyscalTrajectory(fwdfilename)
                for x in traj.nblocks:
                    aseobj = traj[x].to_ase(species=self._get_element_list())
                    fp.append(aseobj.positions)
                    fc.append(list(aseobj.cell))
            if os.path.exists(bkdfilename):
                traj = PyscalTrajectory(bkdfilename)
                for x in traj.nblocks:
                    aseobj = traj[x].to_ase(species=self._get_element_list())
                    bp.append(aseobj.positions)
                    bc.append(list(aseobj.cell))

            fwd_positions.append(fp)
            bkd_positions.append(bp)
            fwd_cells.append(fc)
            bkd_positions.append(bc)

        return fwd_positions, bkd_positions, fwd_cells, bkd_cells

    def collect_output(self):
        self.collect_general_output()
        self.to_hdf()

    def db_entry(self):
        """
        Generate the initial database entry
        Returns:
            (dict): db_dict
        """
        db_dict = super(Calphy, self).db_entry()
        if self.structure:
            if isinstance(self.structure, Atoms):
                parent_structure = self.structure.get_parent_basis()
            else:
                parent_structure = self.structure.copy()
            db_dict["ChemicalFormula"] = parent_structure.get_chemical_formula()

        return db_dict

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf=hdf, group_name=group_name)
        self.input.to_hdf(self.project_hdf5)
        self._output.to_hdf(self.project_hdf5)

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(hdf=hdf, group_name=group_name)
        self.input.from_hdf(self.project_hdf5)
        self._output.from_hdf(self.project_hdf5)
        self._potential_from_hdf()
        # self.create_calc()

    @property
    def publication(self):
        return {
            "calphy": {
                "calphy": {
                    "title": "Automated free-energy calculation from atomistic simulations",
                    "journal": "Physical Review Materials",
                    "volume": "5",
                    "number": "10",
                    "pages": "103801",
                    "year": "2021",
                    "doi": "10.1103/PhysRevMaterials.5.103801",
                    "url": "https://journals.aps.org/prmaterials/abstract/10.1103/PhysRevMaterials.5.103801",
                    "author": [
                        "Menon, Sarath and Lysogorskiy, Yury and Rogal, Jutta and Drautz, Ralf"
                    ],
                }
            }
        }



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np

from pyiron_atomistics.atomistics.job.interactive import GenericInteractive
from pyiron_atomistics.atomistics.structure.atoms import Atoms

try:
    from ase.cell import Cell
except ImportError:
    Cell = None

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class AseJob(GenericInteractive):
    def __init__(self, project, job_name):
        super(AseJob, self).__init__(project, job_name)

        self.__version__ = (
            None  # Reset the version number to the executable is set automatically
        )

    def to_hdf(self, hdf=None, group_name=None):
        super(AseJob, self).to_hdf(hdf=hdf, group_name=group_name)
        self._structure_to_hdf()

    def from_hdf(self, hdf=None, group_name=None):
        super(AseJob, self).from_hdf(hdf=hdf, group_name=group_name)
        self._structure_from_hdf()

    def run_static(self):
        pre_run_mode = self.server.run_mode
        self.server.run_mode.interactive = True
        self.run_if_interactive()
        self.interactive_close()
        self.server.run_mode = pre_run_mode

    def run_if_interactive(self):
        self._ensure_structure_calc_is_set()
        super(AseJob, self).run_if_interactive()
        self.interactive_collect()

    def _ensure_structure_calc_is_set(self):
        if self.structure.calc is None:
            self.set_calculator()

    def set_calculator(self):
        raise NotImplementedError(
            "The set_calculator function is not implemented for this code. Either set "
            "an ase calculator to the structure attribute, or subclass this job and "
            "define set_calculator."
        )

    def interactive_structure_setter(self, structure):
        self.structure.calc.calculate(structure)

    def interactive_positions_setter(self, positions):
        self.structure.positions = positions

    def interactive_initialize_interface(self):
        self.status.running = True
        self._ensure_structure_calc_is_set()
        self.structure.calc.set_label(self.working_directory + "/")
        self._interactive_library = True

    def interactive_close(self):
        if self.interactive_is_activated():
            super(AseJob, self).interactive_close()
            with self.project_hdf5.open("output") as h5:
                if "interactive" in h5.list_groups():
                    for key in h5["interactive"].list_nodes():
                        h5["generic/" + key] = h5["interactive/" + key]

    def interactive_forces_getter(self):
        return self.structure.get_forces()

    def interactive_pressures_getter(self):
        return -self.structure.get_stress(voigt=False)

    def interactive_energy_pot_getter(self):
        return self.structure.get_potential_energy()

    def interactive_energy_tot_getter(self):
        return self.structure.get_potential_energy()

    def interactive_indices_getter(self):
        element_lst = sorted(list(set(self.structure.get_chemical_symbols())))
        return np.array(
            [element_lst.index(el) for el in self.structure.get_chemical_symbols()]
        )

    def interactive_positions_getter(self):
        return self.structure.positions.copy()

    def interactive_steps_getter(self):
        return len(self.interactive_cache[list(self.interactive_cache.keys())[0]])

    def interactive_time_getter(self):
        return self.interactive_steps_getter()

    def interactive_volume_getter(self):
        return self.structure.get_volume()

    def interactive_cells_getter(self):
        return self.structure.cell.copy()

    def write_input(self):
        pass

    def collect_output(self):
        pass

    def run_if_scheduler(self):
        self._create_working_directory()
        super(AseJob, self).run_if_scheduler()

    def interactive_index_organizer(self):
        index_merge_lst = self._interactive_species_lst.tolist() + list(
            np.unique(self._structure_current.get_chemical_symbols())
        )
        el_lst = sorted(set(index_merge_lst), key=index_merge_lst.index)
        current_structure_index = [
            el_lst.index(el) for el in self._structure_current.get_chemical_symbols()
        ]
        previous_structure_index = [
            el_lst.index(el) for el in self._structure_previous.get_chemical_symbols()
        ]
        if not np.array_equal(
            np.array(current_structure_index),
            np.array(previous_structure_index),
        ):
            self._logger.debug("Generic library: indices changed!")
            self.interactive_indices_setter(self._structure_current.indices)

    def _get_structure(self, frame=-1, wrap_atoms=True):
        if (
            self.server.run_mode.interactive
            or self.server.run_mode.interactive_non_modal
        ):
            # Warning: We only copy symbols, positions and cell information - no tags.
            if self.output.indices is not None and len(self.output.indices) != 0:
                indices = self.output.indices[frame]
            else:
                return None
            if len(self._interactive_species_lst) == 0:
                el_lst = list(np.unique(self._structure_current.get_chemical_symbols()))
            else:
                el_lst = self._interactive_species_lst.tolist()
            if indices is not None:
                if wrap_atoms:
                    positions = self.output.positions[frame]
                else:
                    if len(self.output.unwrapped_positions) > max([frame, 0]):
                        positions = self.output.unwrapped_positions[frame]
                    else:
                        positions = (
                            self.output.positions[frame]
                            + self.output.total_displacements[frame]
                        )
                atoms = Atoms(
                    symbols=np.array([el_lst[el] for el in indices]),
                    positions=positions,
                    cell=self.output.cells[frame],
                    pbc=self.structure.pbc,
                )
                # Update indicies to match the indicies in the cache.
                atoms.set_array("indices", indices)
                return atoms
            else:
                return None
        else:
            if (
                self.get("output/generic/cells") is not None
                and len(self.get("output/generic/cells")) != 0
            ):
                return super()._get_structure(frame=frame, wrap_atoms=wrap_atoms)
            else:
                return None



from typing import Optional

import numpy as np


def format_value(v, indent=0):
    if isinstance(v, bool):
        return f" = {v};".lower()
    elif isinstance(v, dict) or isinstance(v, list):
        if len(v) == 0:
            return " {}"
        else:
            return " {\n" + to_sphinx(v, indent + 1) + indent * "\t" + "}"
    else:
        if isinstance(v, np.ndarray):
            v = v.tolist()
        return " = {!s};".format(v)


def to_sphinx(obj, indent=0):
    line = ""
    for k, v in obj.items():
        current_line = indent * "\t" + k.split("___")[0]
        if isinstance(v, list):
            for vv in v:
                line += current_line + format_value(vv, indent) + "\n"
        else:
            line += current_line + format_value(v, indent) + "\n"
    return line


def append_item(group, key, value, n_max=int(1e8)):
    if key not in group:
        group[key] = value
        return group
    else:
        for ii in range(n_max):
            if f"{key}___{ii}" not in group:
                group[f"{key}___{ii}"] = value
                return group
    raise ValueError("Too many items in group")


def fill_values(group=None, **kwargs):
    if group is None:
        group = {}
    for k, v in kwargs.items():
        while k.endswith("_"):
            k = k[:-1]
        if v is not None and v is not False:
            if isinstance(v, list) and len(v) > 0 and isinstance(v[0], dict):
                for i, vv in enumerate(v):
                    group = append_item(group, k, vv)
            else:
                group = append_item(group, k, v)
    return group


def get_all_group(
    structure: Optional[dict] = None,
    basis: Optional[dict] = None,
    pseudo_pot: Optional[dict] = None,
    pw_hamiltonian: Optional[dict] = None,
    paw_pot: Optional[dict] = None,
    paw_hamiltonian: Optional[dict] = None,
    initial_guess: Optional[dict] = None,
    main: Optional[dict] = None,
):
    return fill_values(
        structure=structure,
        basis=basis,
        pseudoPot=pseudo_pot,
        pwHamiltonian=pw_hamiltonian,
        pawPot=paw_pot,
        pawHamiltonian=paw_hamiltonian,
        initialGuess=initial_guess,
        main=main,
    )


def get_structure_group(
    cell: np.ndarray,
    movable: Optional[bool] = None,
    movable_x: Optional[bool] = None,
    movable_y: Optional[bool] = None,
    movable_z: Optional[bool] = None,
    species: Optional[dict | list] = None,
) -> dict:
    """
    Args:
        cell (np.ndarray): Cell matrix
        movable (bool): Allow atoms to move. Default: all atoms are movable,
            unless any movable tag is used for any species/atom.
        movable_x (bool): Allow atoms to move in the x direction.
            Default: movable, unless movableY or movableZ are used.
        movable_y (bool): Allow atoms to move in the y direction.
            Default: movable, unless movableX or movableZ are used.
        movable_z (bool): Allow atoms to move in the z direction.
            Default: movable, unless movableX or movableY are used.
        species (list): Species
    """
    return fill_values(
        cell=cell,
        movable=movable,
        movableX=movable_x,
        movableY=movable_y,
        movableZ=movable_z,
        species=species,
    )


def get_species_group(
    element: str,
    atom: Optional[dict] = None,
) -> dict:
    """
    Args:
        element (str): Element
        atom (list): Atom
    """
    return fill_values(element=element, atom=atom)


def get_atom_group(
    coords: np.ndarray,
    relative: Optional[bool] = None,
    movable_line: Optional[list] = None,
    label: Optional[str] = None,
) -> dict:
    """
        Args:
            coords (np.ndarray): The atomic coordinates as a 3-vector. Un- less the
                relative tag is employed, the coordinates are Cartesian (in bohr)
            relative (bool): The coordinates are given relative to the unit cell vectors.
            movable_line (list): The movement of the atom is restricted to a line.
                The value gives the direction of the line as a 3-vector.
            label (str): Assign a label (or rather a tag) to this atom. If labels
                are used, atoms with different labels are considered inequivalent.
                Think of spin con
    gurations for a use-case.
    """
    return fill_values(
        coords=coords, relative=relative, movableLine=movable_line, label=label
    )


def get_symmetry_group(operator: dict) -> dict:
    """
    Args:
        operator (np.ndarray): Symmetry operator
    """
    return fill_values(operator=operator)


def get_basis_group(
    e_cut: float,
    g_cut: Optional[float] = None,
    folding: Optional[int] = None,
    mesh: Optional[list] = None,
    mesh_accuracy: Optional[float] = None,
    save_memory: Optional[bool] = None,
    k_point: Optional[dict] = None,
) -> dict:
    """
    Args:
        e_cut (float): Energy cutoff
        g_cut (float): Gradient cutoff
        folding (int): Folding
        mesh (list): Mesh
        mesh_accuracy (float): Mesh accuracy
        save_memory (bool): Save memory
        k_point (dict): K point
    """
    return fill_values(
        eCut=e_cut,
        gCut=g_cut,
        folding=folding,
        mesh=mesh,
        meshAccuracy=mesh_accuracy,
        saveMemory=save_memory,
        kPoint=k_point,
    )


def get_k_point_group(
    coords: np.ndarray,
    relative: Optional[bool] = None,
    weight: Optional[float] = None,
) -> dict:
    """
    Args:
        coords (np.ndarray): The k-point coordinates as a 3-vector. Unless the
            relative tag is employed, the coordinates are Cartesian.
        relative (bool): The coordinates are given relative to the unit cell vectors.
        weight (float): The weight of the k-point in the sampling.
    """
    return fill_values(coords=coords, relative=relative, weight=weight)


def get_k_points_group(
    relative: Optional[bool] = None,
    dK: Optional[float] = None,
    from_: Optional[dict | list] = None,
    to: Optional[dict | list] = None,
) -> dict:
    """
    Args:
        relative (bool): The coordinates are given relative to the unit cell
            vectors.
        dK (float): Set the number of intermediate k-points such that the
            distance is at most dK (in 1/bohr).
        from (dict): The from group (within the kPoints group) adds a single
            k-point at the desired position. It may be used multiple times.
        to (dict): The to group (within the kPoints group) adds a line of
            k-points from the previous one to a new position. The number of
            points is set directly with nPoints or indirectly via dK.
    """
    return fill_values(relative=relative, dK=dK, from_=from_, to=to)


def get_k_points_from_group(
    coords: np.ndarray,
    relative: Optional[bool] = None,
    label: Optional[str] = None,
) -> dict:
    """
    Args:
        coords (np.ndarray): The k-point coordinates as a 3-vector. If
            relative flag is not given, the units are 1/bohr.
        relative (bool): The coordinates are given relative to the unit cell
            vectors.
        label (str): Assign a label (or rather a tag) to this k-point. If labels
            are used, k-points with different labels are considered inequivalent.
    """
    return fill_values(coords=coords, relative=relative, label=label)


def get_k_points_to_group(
    coords: np.ndarray,
    relative: Optional[bool] = None,
    label: Optional[str] = None,
    dK: Optional[float] = None,
    n_points: Optional[int] = None,
) -> dict:
    """
    Args:
        coords (np.ndarray): The k-point coordinates as a 3-vector. If
            relative flag is not given, the units are 1/bohr.
        relative (bool): The coordinates are given relative to the unit cell
            vectors.
        label (str): Assign a label (or rather a tag) to this k-point. If labels
            are used, k-points with different labels are considered inequivalent.
        dK (float): Set the number of intermediate k-points such that the
            distance is at most dK (in 1/bohr).
        n_points (int): Specify number of points to add. The final one will be
            at coords.
    """
    return fill_values(
        coords=coords, relative=relative, label=label, dK=dK, nPoints=n_points
    )


def get_paw_pot_group(species: dict) -> dict:
    """
    Args:
        species (dict): Species
    """
    return fill_values(species=species)


def get_paw_pot_species_group(
    potential: str,
    pot_type: str,
    name: Optional[str] = None,
    element: Optional[str] = None,
    l_max_rho: Optional[int] = None,
    angular_grid: Optional[int] = None,
    n_rad_grid: Optional[int] = None,
    check_overlap: Optional[bool] = None,
    r_paw: Optional[float] = None,
) -> dict:
    """
    Args:
        potential (str): Name of the potential file.
        pot_type (str): Type of the potential file, see table below
        name (str): English name of the element
        element (str): Chemical symbol of the element
        l_max_rho (int): Truncate the spherical expansion of densities (and
            compensation charges) at this l.
        angular_grid (int): Choose a different angular grid for xc calculation
            in the PAW sphere. Larger is finer. Default is 7 (110 points).
        n_rad_grid (int): Interpolate to a different radial grid.
        check_overlap (bool): Check that PAW norm is garantueed to be positive
            definite in the limit of large cutoffs. This is on by default.
            Some problematic PAW potentials may fail the check, but work
            normally in some circumstances, so you can switch off the check here.
        r_paw (float): Change the PAW radius used for atomic quantities
            "inside the PAW sphere".

    Potentials:
        - "AbInit": conventional abinit format (non-xml)
        - "AtomPAW": potentials from the atompaw generator (output 2, .atomicdata)
        - "CPPAW": potentials of Blöchl's original PAW code
        - "VASP": vasp potential file
        - "xml": PAW-XML format (abinit, atompaw, gpaw)
    """
    return fill_values(
        potential=potential,
        potType=pot_type,
        name=name,
        element=element,
        lMaxRho=l_max_rho,
        angularGrid=angular_grid,
        nRadGrid=n_rad_grid,
        checkOverlap=check_overlap,
        rPAW=r_paw,
    )


def get_paw_hamiltonian_group(
    xc: str,
    ekt: Optional[float] = None,
    methfessel_paxton: Optional[float] = None,
    fermi_dirac: Optional[float] = None,
    n_empty_states: Optional[int] = None,
    n_excess_electrons: Optional[int] = None,
    spin_polarized: Optional[bool] = None,
    dipole_correction: Optional[bool] = None,
    z_field: Optional[float] = None,
    v_ext: Optional[dict] = None,
    xc_mesh: Optional[dict] = None,
    vdw_correction: Optional[dict] = None,
    hubbard_u: Optional[dict] = None,
) -> dict:
    """
    Args:
        xc (str): The exchange-correlation functional to use.
        ekt (float): The electronic temperature in eV.
        methfessel_paxton (float): If ≥0, use Methfessel-Paxton smearing of
            indicated order. Order 0 is same as Gaussian smearing.
        fermi_dirac (float): If ≥0, use FermiDirac smearing of indicated
            order. Order 0 is the default; order 1 corresponds to first-order
            corrections. Higher orders are not yet implemented.
        n_empty_states (int): The number of empty states to include in the
            calculation.
        n_excess_electrons (int): The number of excess electrons to include in
            the calculation.
        spin_polarized (bool): Whether to perform a spin-polarized calculation.
        dipole_correction (bool): Use the dipole correction for slab systems.
            The in-plane lattice must be perpendicular to the z- axis, and the
            third basis vector must be aligned with the z-axis. For charged
            calculation, this requests the generalized dipole correction,
            which may need some care for initializing the charge (see charged
            in the initialGuess group).
        z_field (float): Use an additional electric field along z when using
            the dipole correction (eV/bohr)
        v_ext (dict): External potential
        xc_mesh (dict): Mesh for the exchange-correlation potential
        vdw_correction (dict): Van der Waals correction
        hubbard_u (dict): Hubbard U
    """
    return fill_values(
        xc=xc,
        ekt=ekt,
        MethfesselPaxton=methfessel_paxton,
        FermiDirac=fermi_dirac,
        nEmptyStates=n_empty_states,
        nExcessElectrons=n_excess_electrons,
        spinPolarized=spin_polarized,
        dipoleCorrection=dipole_correction,
        zField=z_field,
        vExt=v_ext,
        xcMesh=xc_mesh,
        vdwCorrection=vdw_correction,
        HubbardU=hubbard_u,
    )


def get_v_ext_group(file: str) -> dict:
    """
    Args:
        file (str): The file containing the external potential.
    """
    return fill_values(file=file)


def get_xc_mesh_group(
    e_cut: float,
    mesh: Optional[list] = None,
    mesh_accuracy: Optional[float] = None,
) -> dict:
    """
    Args:
        e_cut (float): The energy cutoff for the mesh.
        mesh (list): The mesh for the exchange-correlation potential.
        mesh_accuracy (float): The accuracy of the mesh.
    """
    return fill_values(eCut=e_cut, mesh=mesh, meshAccuracy=mesh_accuracy)


def get_vdw_correction_group(
    method: str,
    combination_rule: Optional[str] = None,
) -> dict:
    """
    Args:
        method (str): The method to use for the van der Waals correction.
        combination_rule (str): The combination rule to use for the van der
            Waals correction.
    """
    return fill_values(method=method, combinationRule=combination_rule)


def get_hubbard_u_group(site: dict | list, verbose: Optional[bool] = None) -> dict:
    """
    Args:
        site (dict): Site
        verbose (bool): Verbose output
    """
    return fill_values(site=site, verbose=verbose)


def get_hubbard_u_site_group(
    element: str,
    species: Optional[int] = None,
    label: Optional[str] = None,
    projector_type: Optional[str] = None,
    l: Optional[int] = None,
    U: Optional[float] = None,
    shift: Optional[float] = None,
) -> dict:
    """
    Args:
        element (str): defines the element via its name
        species (str): defines the element via its species number (1,2,3...)
            within the input file
        label (str): defines the relevant atoms via their label. All atoms
            must belong to the same species. See also label in the atom group
        projector_type (str): use a single projector from the PAW potential.
            Counting starts at 1. l-channel is chosen automatically.
        l (int): use all projectors of a particular l-channel for projection.
            A radial truncation inside the PAW sphere is applied. Use rPAW in
            the pawPot group to change the radius.
        U (float): The Hubbard U value (in eV)
        shift (float): An additional energy shift of the projector (in eV)
    """
    return fill_values(
        element=element,
        species=species,
        label=label,
        projectorType=projector_type,
        l=l,
        U=U,
        shift=shift,
    )


def get_hubbard_u_AO_group(
    orbital: dict,
    element: Optional[str] = None,
    species: Optional[int] = None,
    label: Optional[str] = None,
    n_rad_grid: Optional[int] = None,
    r_cut: Optional[float] = None,
    cut_width: Optional[float] = None,
    U: Optional[float] = None,
    shift: Optional[float] = None,
) -> dict:
    """
    Args:
        orbital (dict): Orbital
        element (str): defines the element via its name
        species (str): defines the element via its species number (1,2,3...)
            within the input file
        label (str): defines the relevant atoms via their label. All atoms
            must belong to the same species. See also label in the atom group
        n_rad_grid (int): number of radial points to represent atomic orbital
            projector. Default: 200
        r_cut (float): The cutoff radius for the atomic orbital (in bohr).
        cut_width (float): The width of the cutoff region (in bohr) Default 0.7 bohr.
        U (float): The Hubbard U value (in eV)
        shift (float): An additional energy shift of the projector (in eV)
    """
    return fill_values(
        orbital=orbital,
        element=element,
        species=species,
        label=label,
        nRadGrid=n_rad_grid,
        rCut=r_cut,
        cutWidth=cut_width,
        U=U,
        shift=shift,
    )


def get_hubbard_u_MO_group(
    orbital: dict,
    element: str,
    species: Optional[int] = None,
    label: Optional[str] = None,
    max_dist: Optional[float] = None,
    min_dist: Optional[float] = None,
    n_interpolate: Optional[int] = None,
    n_rad_grid: Optional[int] = None,
    r_cut: Optional[float] = None,
    cut_width: Optional[float] = None,
    m_MO: Optional[int] = None,
    sign: Optional[int] = None,
    U: Optional[float] = None,
    shift: Optional[float] = None,
) -> dict:
    """
    Args:
        orbital (dict): Orbital
        element (str): defines the element via its name
        species (str): defines the element via its species number (1,2,3...)
            within the input file
        label (str): defines the relevant atoms via their label. All atoms
            must belong to the same species. See also label in the atom group
        max_dist (float): Maximum distance for interpolation
        min_dist (float): Minimum distance for interpolation
        n_interpolate (int): Number of points for interpolation
        n_rad_grid (int): number of radial points to represent atomic orbital
            projector. Default: 200
        r_cut (float): The cutoff radius for the atomic orbital (in bohr).
        cut_width (float): The width of the cutoff region (in bohr) Default 0.7 bohr.
        m_MO (int): The m value of the MO
        sign (int): The sign of the MO
        U (float): The Hubbard U value (in eV)
        shift (float): An additional energy shift of the projector (in eV)
    """
    return fill_values(
        orbital=orbital,
        element=element,
        species=species,
        label=label,
        maxDist=max_dist,
        minDist=min_dist,
        nInterpolate=n_interpolate,
        nRadGrid=n_rad_grid,
        rCut=r_cut,
        cutWidth=cut_width,
        mMO=m_MO,
        sign=sign,
        U=U,
        shift=shift,
    )


def get_orbital_group(
    file: str,
    iot: int,
    from_potential: Optional[bool] = None,
    is_: Optional[int] = None,
) -> dict:
    """
    Args:
        file (str): Get orbital shape from this quamol-type sxb file.
        iot (int): Which orbital to take. Starts at 0.
        from_potential (bool): Get orbital shape from potential. This is not
            recommended.
        is_ (int): species id within file (starts at 0). If not given, assumes
            same species ordering in sxb file as in input file.
    """
    return fill_values(file=file, iot=iot, fromPotential=from_potential, is_=is_)


def get_spin_constraint_group(
    label: Optional[str] = None,
    constraint: Optional[float] = None,
    file: Optional[str] = None,
) -> dict:
    """
    Args:
        label (str): The present constraint applies to atoms with the given label.
        constraint (float): Value of the desired atomic spin
        file (str): Read all spin constraints from this file.
    """
    return fill_values(label=label, constraint=constraint, file=file)


def get_initial_guess_group(
    no_waves_storage: Optional[bool] = None,
    no_rho_storage: Optional[bool] = None,
    waves: Optional[dict] = None,
    rho: Optional[dict] = None,
    occupations: Optional[dict] = None,
    exchange: Optional[dict] = None,
) -> dict:
    """
    Args:
        no_waves_storage (bool): Do not save wavefunctions after the initial
            guess
        no_rho_storage (bool): Do not save the density after the initial
            guess.
        waves (dict): Waves
        rho (dict): Rho
        occupations (dict): Occupations
        exchange (dict): Exchange
    """
    return fill_values(
        noWavesStorage=no_waves_storage,
        noRhoStorage=no_rho_storage,
        waves=waves,
        rho=rho,
        occupations=occupations,
        exchange=exchange,
    )


def get_waves_group(
    file: Optional[str] = None,
    random: Optional[bool] = None,
    keep_waves_on_disk: Optional[bool] = None,
    lcao: Optional[dict] = None,
) -> dict:
    """
    Args:
        file (str): Read wave functions from this file.
        random (bool): Initialize wave functions randomly.
        keep_waves_on_disk (bool): Keep waves on disk, load only a single
            k-point at each time. May save a lot of RAM, but can be quite a
            bottleneck on small systems.
        lcao (dict): LCAO
    """
    return fill_values(
        file=file, random=random, keepWavesOnDisk=keep_waves_on_disk, lcao=lcao
    )


def get_lcao_group(
    max_steps: Optional[int] = None,
    d_energy: Optional[float] = None,
) -> dict:
    """
    The lcao group within the initialGuess.waves group finetunes the LCAO
    calculation, if necessary. Notably, you can iterate the LCAO calculation to
    self-consistency. This is generally no good idea.

    Args:
        max_steps (int): Maximum number of SCF steps
        d_energy (float): Energy convergence criterion
    """
    return fill_values(maxSteps=max_steps, dEnergy=d_energy)


def get_rho_group(
    file: Optional[str] = None,
    from_wave: Optional[bool] = None,
    random: Optional[bool] = None,
    atomic_orbitals: Optional[bool] = None,
    spin_moment: Optional[bool] = None,
    atomic_spin: Optional[dict] = None,
    charged: Optional[dict] = None,
) -> dict:
    """
    Args:
        file (str): Read density from this file.
        from_wave (bool): Compute from the wave functions (which must be from
            file in this case).
        random (bool): Initialize density randomly.
        atomic_orbitals (bool): Initialize density from atomic orbitals.
        spin_moment (bool): When from atomic densities, apply a global spin
            polarization.
        atomic_spin (dict): Atomic spin
        charged (dict): Charged
    """
    return fill_values(
        file=file,
        fromWave=from_wave,
        random=random,
        atomicOrbitals=atomic_orbitals,
        spinMoment=spin_moment,
        atomicSpin=atomic_spin,
        charged=charged,
    )


def get_atomic_spin_group(
    spin: Optional[float] = None,
    label: Optional[str] = None,
    file: Optional[str] = None,
) -> dict:
    """
    Args:
        spin (float): The spin of the atom.
        label (str): For which atoms to apply this spin
        file (str): Read atomic spins from this file (one spin per line), one
            per atom, in sequential order.
    """
    return fill_values(spin=spin, label=label, file=file)


def get_charged_group(
    charge: float,
    beta: Optional[float] = None,
    z: Optional[float] = None,
    coords: Optional[np.ndarray] = None,
) -> dict:
    """
    Args:
        charge (float): The classical charge (i.e. -nExcessElectrons from the
            PAWHamiltonian or PWHamiltonian group).
        beta (float): Gaussian broadening
        z (float): Request a sheet charge at this z
        coords (np.ndarray): Request a Gaussian charge at this position
            (in bohr).
    """
    return fill_values(charge=charge, beta=beta, z=z, coords=coords)


def get_occupations_group(
    values: Optional[list] = None,
    range: Optional[list] = None,
) -> dict:
    """
    Args:
        values (list): The occupation numbers for each orbital.
        range (list): The range of occupation numbers.
    """
    return fill_values(values=values, range=range)


def get_exchange_group(file: str) -> dict:
    """
    Args:
        file (str): Read exchange from this file.
    """
    return fill_values(file=file)


def get_pseudo_pot_group(species: dict) -> dict:
    """
    Args:
        species (dict): Species
    """
    return fill_values(species=species)


def get_pseudo_pot_species_group(
    name: str,
    potential: str,
    valence_charge: float,
    l_max: str,
    l_loc: str,
    lcao_orbitals: list,
    r_gauss: float,
    reciprocal_mass: float,
    damping_mass: float,
    ionic_mass: float,
    element: Optional[str] = None,
) -> dict:
    """
    Args:
        name (str): The name of the species
        potential (str): The potential file
        valence_charge (float): The valence charge
        l_max (str): The maximum l value
        l_loc (str): The l value for the local part
        lcao_orbitals (list): The LCAO orbitals
        r_gauss (float): The Gaussian radius
        reciprocal_mass (float): The reciprocal mass
        damping_mass (float): The damping mass
        ionic_mass (float): The ionic mass
        element (str): The element
    """
    return fill_values(
        name=name,
        potential=potential,
        valenceCharge=valence_charge,
        lMax=l_max,
        lLoc=l_loc,
        lcaoOrbitals=lcao_orbitals,
        rGauss=r_gauss,
        reciprocalMass=reciprocal_mass,
        dampingMass=damping_mass,
        ionicMass=ionic_mass,
        element=element,
    )


def get_PW_hamiltonian_group(
    xc: str,
    ekt: Optional[float] = None,
    methfessel_paxton: Optional[float] = None,
    fermi_dirac: Optional[float] = None,
    n_empty_states: Optional[int] = None,
    n_excess_electrons: Optional[int] = None,
    spin_polarized: Optional[bool] = None,
    dipole_correction: Optional[bool] = None,
    z_field: Optional[float] = None,
) -> dict:
    """
    Args:
        xc (str): The exchange-correlation functional to use.
        ekt (float): The electronic temperature in eV.
        methfessel_paxton (float): If ≥0, use Methfessel-Paxton smearing of
            indicated order. Order 0 is same as Gaussian smearing.
        fermi_dirac (float): If ≥0, use FermiDirac smearing of indicated
            order. Order 0 is the default; order 1 corresponds to first-order
            corrections. Higher orders are not yet implemented.
        n_empty_states (int): The number of empty states to include in the
            calculation.
        n_excess_electrons (int): The number of excess electrons to include in
            the calculation.
        spin_polarized (bool): Whether to perform a spin-polarized calculation.
        dipole_correction (bool): Use the dipole correction for slab systems.
            The in-plane lattice must be perpendicular to the z- axis, and the
            third basis vector must be aligned with the z-axis. For charged
            calculation, this requests the generalized dipole correction,
            which may need some care for initializing the charge (see charged
            in the initialGuess group).
        z_field (float): Use an additional electric field along z when using
            the dipole correction (eV/bohr)
    """
    return fill_values(
        xc=xc,
        ekt=ekt,
        MethfesselPaxton=methfessel_paxton,
        FermiDirac=fermi_dirac,
        nEmptyStates=n_empty_states,
        nExcessElectrons=n_excess_electrons,
        spinPolarized=spin_polarized,
        dipoleCorrection=dipole_correction,
        zField=z_field,
    )


def get_main_group(**kwargs) -> dict:
    """
    Args:
        steps (dict): Steps
    """
    return fill_values(**kwargs)


def get_CCG_group(
    d_energy: Optional[float] = None,
    max_steps: Optional[int] = None,
    print_steps: Optional[int] = None,
    initial_diag: Optional[bool] = None,
    final_diag: Optional[bool] = None,
    kappa: Optional[float] = None,
    keep_occ_fixed: Optional[bool] = None,
    ekt: Optional[float] = None,
    dipole_correction: Optional[bool] = None,
    no_rho_storage: Optional[bool] = None,
    no_wave_storage: Optional[bool] = None,
) -> dict:
    """
    Args:
        d_energy (float): Energy convergence criterion
        max_steps (int): Maximum number of SCF steps
        print_steps (int): Print SCF steps
        initial_diag (bool): Initial diagonalization
        final_diag (bool): Final diagonalization
        kappa (float): Kappa parameter
        keep_occ_fixed (bool): Keep occupation fixed
        ekt (float): Temperature
        dipole_correction (bool): Dipole correction
        no_rho_storage (bool): Do not store density
        no_wave_storage (bool): Do not store wave functions
    """
    return fill_values(
        dEnergy=d_energy,
        maxSteps=max_steps,
        printSteps=print_steps,
        initialDiag=initial_diag,
        finalDiag=final_diag,
        kappa=kappa,
        keepOccFixed=keep_occ_fixed,
        ekt=ekt,
        dipoleCorrection=dipole_correction,
        noRhoStorage=no_rho_storage,
        noWaveStorage=no_wave_storage,
    )


def get_scf_CCG_group(
    d_rel_eps: Optional[float] = None,
    max_steps_CCG: Optional[int] = None,
    d_energy: Optional[float] = None,
) -> dict:
    """
    Args:
        d_rel_eps (float): Relative energy convergence criterion
        max_steps_CCG (int): Maximum number of CCG steps
        d_energy (float): Energy convergence criterion
    """
    return fill_values(
        dRelEps=d_rel_eps,
        maxStepsCCG=max_steps_CCG,
        dEnergy=d_energy,
    )


def get_scf_block_CCG_group(
    d_rel_eps: Optional[float] = None,
    max_steps_CCG: Optional[int] = None,
    block_size: Optional[int] = None,
    n_sloppy: Optional[int] = None,
    d_energy: Optional[float] = None,
    verbose: Optional[bool] = None,
    numerical_limit: Optional[bool] = None,
) -> dict:
    """
    Args:
        d_rel_eps (float): Relative energy convergence criterion
        max_steps_CCG (int): Maximum number of CCG steps
        block_size (int): Block size
        n_sloppy (int): Number of sloppy steps
        d_energy (float): Energy convergence criterion
        verbose (bool): Verbose output
        numerical_limit (bool): Numerical limit
    """
    return fill_values(
        dRelEps=d_rel_eps,
        maxStepsCCG=max_steps_CCG,
        blockSize=block_size,
        nSloppy=n_sloppy,
        dEnergy=d_energy,
        verbose=verbose,
        numericalLimit=numerical_limit,
    )


def get_preconditioner_group(
    type: Optional[str] = "KERKER",
    scaling: Optional[float] = None,
    spin_scaling: Optional[float] = None,
    kerker_camping: Optional[float] = None,
    dielec_constant: Optional[float] = None,
) -> dict:
    return fill_values(
        type=type,
        scaling=scaling,
        spinScaling=spin_scaling,
        kerkerCamping=kerker_camping,
        dielecConstant=dielec_constant,
    )


def get_scf_diag_group(
    d_energy: Optional[float] = None,
    max_steps: Optional[int] = None,
    max_residue: Optional[float] = None,
    print_steps: Optional[int] = None,
    mixing_method: Optional[str] = None,
    n_pulay_steps: Optional[int] = None,
    rho_mixing: Optional[float] = None,
    spin_mixing: Optional[float] = None,
    keep_rho_fixed: Optional[bool] = None,
    keep_occ_fixed: Optional[bool] = None,
    keep_spin_fixed: Optional[bool] = None,
    ekt: Optional[float] = None,
    dipole_correction: Optional[bool] = None,
    d_spin_moment: Optional[float] = None,
    no_rho_storage: Optional[bool] = None,
    no_wave_storage: Optional[bool] = None,
    CCG: Optional[dict] = None,
    block_CCG: Optional[dict] = None,
    preconditioner: Optional[dict] = None,
) -> dict:
    """
    Args:
        d_energy (float): Energy convergence criterion
        max_steps (int): Maximum number of SCF steps
        max_residue (float): Residue convergence criterion
        print_steps (int): Print SCF steps
        mixing_method (str): Mixing method
        n_pulay_steps (int): Number of Pulay steps
        rho_mixing (float): Density mixing parameter
        spin_mixing (float): Spin mixing parameter
        keep_rho_fixed (bool): Keep density fixed
        keep_occ_fixed (bool): Keep occupation fixed
        keep_spin_fixed (bool): Keep spin fixed
        ekt (float): Temperature
        dipole_correction (bool): Dipole correction
        d_spin_moment (float): Spin moment convergence criterion
        no_rho_storage (bool): Do not store density
        no_wave_storage (bool): Do not store wave functions
        CCG (dict): Conjugate gradient method
        block_CCG (dict): Block conjugate gradient method
        preconditioner (dict): Preconditioner
    """
    return fill_values(
        dEnergy=d_energy,
        maxSteps=max_steps,
        maxResidue=max_residue,
        printSteps=print_steps,
        mixingMethod=mixing_method,
        nPulaySteps=n_pulay_steps,
        rhoMixing=rho_mixing,
        spinMixing=spin_mixing,
        keepRhoFixed=keep_rho_fixed,
        keepOccFixed=keep_occ_fixed,
        keepSpinFixed=keep_spin_fixed,
        ekt=ekt,
        dipoleCorrection=dipole_correction,
        dSpinMoment=d_spin_moment,
        noRhoStorage=no_rho_storage,
        noWaveStorage=no_wave_storage,
        CCG=CCG,
        blockCCG=block_CCG,
        preconditioner=preconditioner,
    )


def get_born_oppenheimer_group(
    scf_diag: Optional[dict] = None,
) -> dict:
    """
    Args:
        scf_diag (dict): SCF diagonalization
    """
    return fill_values(scfDiag=scf_diag)


def get_QN_group(
    max_steps: Optional[int] = None,
    dX: Optional[float] = None,
    dF: Optional[float] = None,
    d_energy: Optional[float] = None,
    max_step_length: Optional[float] = None,
    hessian: Optional[str] = None,
    drift_filter: Optional[bool] = None,
    born_oppenheimer: Optional[dict] = None,
) -> dict:
    """
    Args:
        max_steps (int): Maximum number of steps
        dX (float): Position convergence criterion
        dF (float): Force convergence criterion
        d_energy (float): Energy convergence criterion
        max_step_length (float): Maximum step length
        hessian (str): Initialize Hessian from file
        drift_filter (bool): Drift filter
        born_oppenheimer (dict): Born-Oppenheimer
    """
    return fill_values(
        maxSteps=max_steps,
        dX=dX,
        dF=dF,
        dEnergy=d_energy,
        maxStepLength=max_step_length,
        hessian=hessian,
        driftFilter=drift_filter,
        bornOppenheimer=born_oppenheimer,
    )


def get_linQN_group(
    max_steps: Optional[int] = None,
    dX: Optional[float] = None,
    dF: Optional[float] = None,
    d_energy: Optional[float] = None,
    max_step_length: Optional[float] = None,
    n_projectors: Optional[int] = None,
    hessian: Optional[str] = None,
    drift_filter: Optional[bool] = None,
    born_oppenheimer: Optional[dict] = None,
) -> dict:
    """
    Args:
        max_steps (int): Maximum number of steps
        dX (float): Position convergence criterion
        dF (float): Force convergence criterion
        d_energy (float): Energy convergence criterion
        max_step_length (float): Maximum step length
        n_projectors (int): Number of projectors
        hessian (str): Initialize Hessian from file
        drift_filter (bool): Drift filter
        born_oppenheimer (dict): Born-Oppenheimer
    """
    return fill_values(
        maxSteps=max_steps,
        dX=dX,
        dF=dF,
        dEnergy=d_energy,
        maxStepLength=max_step_length,
        nProjectors=n_projectors,
        hessian=hessian,
        driftFilter=drift_filter,
        bornOppenheimer=born_oppenheimer,
    )


def get_ricQN_group(
    max_steps: Optional[int] = None,
    dX: Optional[float] = None,
    dF: Optional[float] = None,
    d_energy: Optional[float] = None,
    max_step_length: Optional[float] = None,
    n_projectors: Optional[int] = None,
    soft_mode_damping: Optional[float] = None,
    drift_filter: Optional[bool] = None,
    born_oppenheimer: Optional[dict] = None,
) -> dict:
    """
    Args:
        max_steps (int): Maximum number of steps
        dX (float): Position convergence criterion
        dF (float): Force convergence criterion
        d_energy (float): Energy convergence criterion
        max_step_length (float): Maximum step length
        n_projectors (int): Number of projectors
        soft_mode_damping (float): Soft mode damping
        drift_filter (bool): Drift filter
        born_oppenheimer (dict): Born-Oppenheimer
    """
    return fill_values(
        maxSteps=max_steps,
        dX=dX,
        dF=dF,
        dEnergy=d_energy,
        maxStepLength=max_step_length,
        nProjectors=n_projectors,
        softModeDamping=soft_mode_damping,
        driftFilter=drift_filter,
        bornOppenheimer=born_oppenheimer,
    )


def get_ric_group(
    max_dist: Optional[float] = None,
    typify_threshold: Optional[float] = None,
    rms_threshold: Optional[float] = None,
    plane_cut_limit: Optional[float] = None,
    with_angles: Optional[bool] = None,
    bvk_atoms: Optional[str] = None,
    born_oppenheimer: Optional[dict] = None,
) -> dict:
    """
    Args:
        max_dist (float): Maximum distance
        typify_threshold (float): Typify threshold
        rms_threshold (float): RMS threshold
        plane_cut_limit (float): Plane cut limit
        with_angles (bool): With angles
        bvk_atoms (str): (experimental) List of atom ids (starting from 1) for
            which born-von-Karman transversal force constants are added. The
            comma-separated list must be enclosed by square brackets []. This
            adds a bond-directional coordinate to each bond of the atoms in the
            list.
        born_oppenheimer (dict): Born-Oppenheimer
    """
    return fill_values(
        maxDist=max_dist,
        typifyThreshold=typify_threshold,
        rmsThreshold=rms_threshold,
        planeCutLimit=plane_cut_limit,
        withAngles=with_angles,
        bvkAtoms=bvk_atoms,
        bornOppenheimer=born_oppenheimer,
    )


def get_ricTS_group(
    max_steps: Optional[int] = None,
    dX: Optional[float] = None,
    dF: Optional[float] = None,
    d_energy: Optional[float] = None,
    n_projectors: Optional[int] = None,
    max_step_length: Optional[float] = None,
    trans_curvature: Optional[float] = None,
    any_stationary_point: Optional[bool] = None,
    max_dir_rot: Optional[float] = None,
    scheme: Optional[int] = None,
    drift_filter: Optional[bool] = None,
    born_oppenheimer: Optional[dict] = None,
) -> dict:
    """
    Args:
        max_steps (int): Maximum number of steps
        dX (float): Position convergence criterion
        dF (float): Force convergence criterion
        d_energy (float): Energy convergence criterion
        n_projectors (int): Number of projectors
        max_step_length (float): Maximum step length
        trans_curvature (float): Transversal curvature
        any_stationary_point (bool): Any stationary point
        max_dir_rot (float): Maximum direction rotation
        scheme (int): Scheme
        drift_filter (bool): Drift filter
        born_oppenheimer (dict): Born-Oppenheimer
    """
    return fill_values(
        maxSteps=max_steps,
        dX=dX,
        dF=dF,
        dEnergy=d_energy,
        nProjectors=n_projectors,
        maxStepLength=max_step_length,
        transCurvature=trans_curvature,
        anyStationaryPoint=any_stationary_point,
        maxDirRot=max_dir_rot,
        scheme=scheme,
        driftFilter=drift_filter,
        bornOppenheimer=born_oppenheimer,
    )



import re
import warnings
from functools import cached_property
from pathlib import Path

import h5py
import numpy as np
import scipy.constants

BOHR_TO_ANGSTROM = (
    scipy.constants.physical_constants["Bohr radius"][0] / scipy.constants.angstrom
)
HARTREE_TO_EV = scipy.constants.physical_constants["Hartree energy in eV"][0]
HARTREE_OVER_BOHR_TO_EV_OVER_ANGSTROM = HARTREE_TO_EV / BOHR_TO_ANGSTROM


def splitter(arr, counter):
    if len(arr) == 0 or len(counter) == 0:
        return []
    arr_new = []
    spl_loc = list(np.where(np.array(counter) == min(counter))[0])
    spl_loc.append(None)
    for ii, ll in enumerate(spl_loc[:-1]):
        arr_new.append(np.array(arr[ll : spl_loc[ii + 1]]).tolist())
    return arr_new


def collect_energy_dat(file_name="energy.dat", cwd=None):
    """

    Args:
        file_name (str): file name
        cwd (str): directory path

    Returns:
        (dict): results

    """
    path = Path(file_name)
    if cwd is not None:
        path = Path(cwd) / path
    energies = np.loadtxt(str(path), ndmin=2)
    results = {"scf_computation_time": splitter(energies[:, 1], energies[:, 0])}
    results["scf_energy_int"] = splitter(energies[:, 2] * HARTREE_TO_EV, energies[:, 0])

    def en_split(e, counter=energies[:, 0]):
        return splitter(e * HARTREE_TO_EV, counter)

    if len(energies[0]) == 7:
        results["scf_energy_free"] = en_split(energies[:, 3])
        results["scf_energy_zero"] = en_split(energies[:, 4])
        results["scf_energy_band"] = en_split(energies[:, 5])
        results["scf_electronic_entropy"] = en_split(energies[:, 6])
    else:
        results["scf_energy_band"] = en_split(energies[:, 3])
    return results


def collect_residue_dat(file_name="residue.dat", cwd="."):
    """

    Args:
        file_name (str): file name
        cwd (str): directory path

    Returns:
        (dict): results

    """
    if cwd is None:
        cwd = "."
    residue = np.loadtxt(str(Path(cwd) / Path(file_name)), ndmin=2)
    if len(residue) == 0:
        return {}
    return {"scf_residue": splitter(residue[:, 1:].squeeze(), residue[:, 0])}


def _collect_eps_dat(file_name="eps.dat", cwd=None):
    """

    Args:
        file_name:
        cwd:

    Returns:

    """
    path = Path(file_name)
    if cwd is not None:
        path = Path(cwd) / path
    return np.loadtxt(str(path), ndmin=2)[..., 1:]


def collect_eps_dat(file_name=None, cwd=None, spins=True):
    if file_name is not None:
        values = [_collect_eps_dat(file_name=file_name, cwd=cwd)]
    elif spins:
        values = [_collect_eps_dat(file_name=f"eps.{i}.dat", cwd=cwd) for i in [0, 1]]
    else:
        values = [_collect_eps_dat(file_name="eps.dat", cwd=cwd)]
    values = np.stack(values, axis=0)
    return {"bands_eigen_values": values.reshape((-1,) + values.shape)}


def collect_energy_struct(file_name="energy-structOpt.dat", cwd=None):
    """

    Args:
        file_name (str): file name
        cwd (str): directory path

    Returns:
        (dict): results

    """
    path = Path(file_name)
    if cwd is not None:
        path = Path(cwd) / path
    return {
        "energy_free": np.loadtxt(str(path), ndmin=2).reshape(-1, 2)[:, 1]
        * HARTREE_TO_EV
    }


def check_permutation(index_permutation):
    if index_permutation is None:
        return
    indices, counter = np.unique(index_permutation, return_counts=True)
    if np.any(counter != 1):
        raise ValueError("multiple entries in the index_permutation")
    if np.any(np.diff(np.sort(indices)) != 1):
        raise ValueError("missing entries in the index_permutation")


def collect_spins_dat(file_name="spins.dat", cwd=None, index_permutation=None):
    """

    Args:
        file_name (str): file name
        cwd (str): directory path
        index_permutation (numpy.ndarray): Indices for the permutation

    Returns:
        (dict): results

    """
    check_permutation(index_permutation)
    path = Path(file_name)
    if cwd is not None:
        path = Path(cwd) / path
    spins = np.loadtxt(str(path), ndmin=2)
    if index_permutation is not None:
        s = np.array([ss[index_permutation] for ss in spins[:, 1:]])
    else:
        s = spins[:, 1:]
    return {"atom_scf_spins": splitter(s, spins[:, 0])}


def collect_relaxed_hist(file_name="relaxHist.sx", cwd=None, index_permutation=None):
    """

    Args:
        file_name (str): file name
        cwd (str): directory path
        index_permutation (numpy.ndarray): Indices for the permutation

    Returns:
        (dict): results

    # TODO: parse movable, elements, species etc.
    """
    check_permutation(index_permutation)
    path = Path(file_name)
    if cwd is not None:
        path = Path(cwd) / path
    with open(str(path), "r") as f:
        file_content = "".join(f.readlines())
    n_steps = max(len(re.findall("// --- step \d", file_content, re.MULTILINE)), 1)
    f_v = ",".join(3 * [r"\s*([\d.-]+)"])

    def get_value(term, f=file_content, n=n_steps, p=index_permutation):
        value = (
            np.array(re.findall(term, f, re.MULTILINE)).astype(float).reshape(n, -1, 3)
        )
        if p is not None:
            value = np.array([ff[p] for ff in value])
        return value

    cell = re.findall(
        r"cell = \[\[" + r"\],\n\s*\[".join(3 * [f_v]) + r"\]\];",
        file_content,
        re.MULTILINE,
    )
    cell = np.array(cell).astype(float).reshape(n_steps, 3, 3) * BOHR_TO_ANGSTROM
    return {
        "positions": get_value(r"atom {coords = \[" + f_v + r"\];") * BOHR_TO_ANGSTROM,
        "forces": get_value(r"force  = \[" + f_v + r"\]; }"),
        "cell": cell,
    }


class SphinxLogParser:
    def __init__(self, file_name="sphinx.log", cwd=None, index_permutation=None):
        """
        Args:
            file_name (str): file name
            cwd (str): directory path
            index_permutation (numpy.ndarray): Indices for the permutation

        """
        path = Path(file_name)
        if cwd is not None:
            path = Path(cwd) / path
        with open(str(path), "r") as sphinx_log_file:
            self.log_file = sphinx_log_file.read()
        self._scf_not_entered = False
        self._check_enter_scf()
        self._log_main = None
        self._n_atoms = None
        check_permutation(index_permutation)
        self._index_permutation = index_permutation
        self.generic_dict = {
            "volume": self.get_volume,
            "forces": self.get_forces,
            "job_finished": self.job_finished,
        }
        self.dft_dict = {
            "n_valence": self.get_n_valence,
            "bands_k_weights": self.get_bands_k_weights,
            "kpoints_cartesian": self.get_kpoints_cartesian,
            "bands_e_fermi": self.get_fermi,
            "bands_occ": self.get_occupancy,
            "bands_eigen_values": self.get_band_energy,
            "scf_convergence": self.get_convergence,
            "scf_energy_int": self.get_energy_int,
            "scf_energy_free": self.get_energy_free,
            "scf_magnetic_forces": self.get_magnetic_forces,
        }

    @property
    def index_permutation(self):
        return self._index_permutation

    @property
    def spin_enabled(self):
        return len(re.findall("Spin moment:", self.log_file)) > 0

    @property
    def log_main(self):
        if self._log_main is None:
            match = re.search("Enter Main Loop", self.log_file)
            self._log_main = match.end() + 1
        return self.log_file[self._log_main :]

    def job_finished(self):
        if (
            len(re.findall("Program exited normally.", self.log_file, re.MULTILINE))
            == 0
        ):
            warnings.warn("scf loops did not converge")
            return False
        return True

    def _check_enter_scf(self):
        if len(re.findall("Enter Main Loop", self.log_file, re.MULTILINE)) == 0:
            warnings.warn("Log file created but first scf loop not reached")
            self._scf_not_entered = True

    def get_n_valence(self):
        log = self.log_file.split("\n")
        return {
            log[ii - 1].split()[1]: int(ll.split("=")[-1])
            for ii, ll in enumerate(log)
            if ll.startswith("| Z=")
        }

    @property
    def _log_k_points(self):
        start_match = re.search(
            "-ik-     -x-      -y-       -z-    \|  -weight-    -nG-    -label-",
            self.log_file,
        )
        log_part = self.log_file[start_match.end() + 1 :]
        log_part = log_part[: re.search("^\n", log_part, re.MULTILINE).start()]
        return log_part.split("\n")[:-2]

    def get_bands_k_weights(self):
        return np.array([float(kk.split()[6]) for kk in self._log_k_points])

    @property
    def _rec_cell(self):
        log_extract = re.findall("b[1-3]:.*$", self.log_file, re.MULTILINE)
        return (
            np.array([ll.split()[1:4] for ll in log_extract]).astype(float)
            / BOHR_TO_ANGSTROM
        )[:3]

    def get_kpoints_cartesian(self):
        return np.einsum("ni,ij->nj", self.k_points, self._rec_cell)

    @property
    def k_points(self):
        return np.array(
            [[float(kk.split()[i]) for i in range(2, 5)] for kk in self._log_k_points]
        )

    def get_volume(self):
        volume = re.findall("Omega:.*$", self.log_file, re.MULTILINE)
        if len(volume) > 0:
            volume = float(volume[0].split()[1])
            volume *= BOHR_TO_ANGSTROM**3
        else:
            volume = 0
        return np.array(self.n_steps * [volume])

    @property
    def counter(self):
        return [
            int(re.sub("[^0-9]", "", line.split("=")[0]))
            for line in re.findall("F\(.*$", self.log_main, re.MULTILINE)
        ]

    def _get_energy(self, pattern):
        c, F = np.array(re.findall(pattern, self.log_main, re.MULTILINE)).T
        return splitter(F.astype(float) * HARTREE_TO_EV, c.astype(int))

    def get_energy_free(self):
        return self._get_energy(pattern=r"F\((\d+)\)=(-?\d+\.\d+)")

    def get_energy_int(self):
        return self._get_energy(pattern=r"eTot\((\d+)\)=(-?\d+\.\d+)")

    @property
    def n_atoms(self):
        if self._n_atoms is None:
            self._n_atoms = len(
                np.unique(re.findall("^Species.*\{", self.log_main, re.MULTILINE))
            )
        return self._n_atoms

    def get_forces(self):
        """
        Returns:
            (numpy.ndarray): Forces of the shape (n_steps, n_atoms, 3)
        """
        str_fl = "([-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?)"
        pattern = r"Atom: (\d+)\t{" + ",".join(3 * [str_fl]) + r"\}"
        arr = np.array(re.findall(pattern, self.log_file))
        if len(arr) == 0:
            return []
        forces = arr[:, 1:].astype(float).reshape(-1, self.n_atoms, 3)
        forces *= HARTREE_OVER_BOHR_TO_EV_OVER_ANGSTROM
        if self.index_permutation is not None:
            for ii, ff in enumerate(forces):
                forces[ii] = ff[self.index_permutation]
        return forces

    def get_magnetic_forces(self):
        """
        Returns:
            (numpy.ndarray): Magnetic forces of the shape (n_steps, n_atoms)
        """
        magnetic_forces = [
            HARTREE_TO_EV * float(line.split()[-1])
            for line in re.findall("^nu\(.*$", self.log_main, re.MULTILINE)
        ]
        if len(magnetic_forces) != 0:
            magnetic_forces = np.array(magnetic_forces).reshape(-1, self.n_atoms)
            if self.index_permutation is not None:
                for ii, mm in enumerate(magnetic_forces):
                    magnetic_forces[ii] = mm[self.index_permutation]
        return splitter(magnetic_forces, self.counter)

    @property
    def n_steps(self):
        return len(re.findall("\| SCF calculation", self.log_file, re.MULTILINE))

    def _parse_band(self, term):
        content = re.findall(term, self.log_main, re.MULTILINE)
        if len(content) == 0:
            return []
        arr = np.loadtxt(content, ndmin=2)
        shape = (-1, len(self.k_points), arr.shape[-1])
        if self.spin_enabled:
            shape = (-1, 2, len(self.k_points), shape[-1])
        return arr.reshape(shape)

    def get_band_energy(self):
        return self._parse_band("final eig \[eV\]:(.*)$")

    def get_occupancy(self):
        return self._parse_band("final focc:(.*)$")

    def get_convergence(self):
        conv_dict = {
            "WARNING: Maximum number of steps exceeded": False,
            "Convergence reached.": True,
        }
        key = "|".join(list(conv_dict.keys())).replace(".", "\.")
        items = re.findall(key, self.log_main, re.MULTILINE)
        convergence = [conv_dict[k] for k in items]
        diff = self.n_steps - len(convergence)
        for _ in range(diff):
            convergence.append(False)
        return convergence

    def get_fermi(self):
        pattern = r"Fermi energy:\s+(-?\d+\.\d+)\s+eV"
        return np.array(re.findall(pattern, self.log_main)).astype(float)

    @property
    def results(self):
        if self._scf_not_entered:
            return {}
        results = {"generic": {}, "dft": {}}
        for key, func in self.generic_dict.items():
            value = func()
            if key == "job_finished" or len(value) > 0:
                results["generic"][key] = value
        for key, func in self.dft_dict.items():
            value = func()
            if len(value) > 0:
                results["dft"][key] = value
        return results


class SphinxWavesReader:
    """Class to read SPHInX waves.sxb files (HDF5 format)

    Initialize with waves.sxb filename, or use load ()

    """

    def __init__(self, file_name="waves.sxb", cwd="."):
        """
        Args:
            file_name (str): file name
            cwd (str): directory path
        """
        if Path(file_name).is_absolute():
            self.wfile = h5py.File(Path(file_name))
        else:
            path = Path(cwd) / Path(file_name)
            self.wfile = h5py.File(path)

    @property
    def _n_gk(self):
        return self.wfile["nGk"][:]

    @cached_property
    def mesh(self):
        return self.wfile["meshDim"][:]

    @property
    def Nx(self):
        return self.mesh[0]

    @property
    def Ny(self):
        return self.mesh[1]

    @property
    def Nz(self):
        return self.mesh[2]

    @cached_property
    def n_states(self):
        return self.wfile["nPerK"][0]

    @cached_property
    def n_spin(self):
        return self.wfile["nSpin"].shape[0]

    @cached_property
    def k_weights(self):
        return self.wfile["kWeights"][:]

    @cached_property
    def k_vec(self):
        return self.wfile["kVec"][:]

    @cached_property
    def eps(self):
        """All eigenvalues (in Hartree) as (nk,n_states) block"""
        return (self.wfile["eps"][:].reshape(-1, self.n_spin, self.n_states)).T

    @cached_property
    def _fft_idx(self):
        fft_idx = []
        off = 0
        for ngk in self._n_gk:
            fft_idx.append(self.wfile["fftIdx"][off : off + ngk])
            off += ngk
        return fft_idx

    def get_psi_rec(self, i, ispin, ik, compact=False):
        """
        Loads a single wavefunction on full FFT mesh of shape mesh.

        params: i: state index (int)
               ispin: spin index (int)
               ik: k index (int)
               compact: (bool)
        returns:
            res: complex valued wavefunction indexed by (i,ispin,ik) loaded on to the FFT mesh.
            compact_wave: compact wavefunctions, without loading on FFT mesh.
        """
        # translate indices to pythonic style.
        i = np.arange(self.n_states)[i]
        ik = np.arange(self.nk)[ik]
        ispin = np.arange(self.n_spin)[ispin]

        off = self._n_gk[ik] * (i + ispin * self.n_states)
        psire = self.wfile[f"psi-{ik+1}.re"][off : off + self._n_gk[ik]]
        psiim = self.wfile[f"psi-{ik+1}.im"][off : off + self._n_gk[ik]]
        compact_wave = psire + 1j * psiim
        if compact:
            return compact_wave
        res = np.zeros(shape=self.mesh, dtype=np.complex128)
        res.flat[self._fft_idx[ik]] = compact_wave
        return res

    @property
    def nk(self):
        """Number of k-points"""
        return self.k_weights.shape[0]



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os
import warnings
from collections import OrderedDict

import numpy as np

from pyiron_atomistics.atomistics.structure.atoms import Atoms

__author__ = "Sudarsan Surendralal"
__copyright__ = (
    "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


def read_atoms(
    filename="CONTCAR",
    return_velocities=False,
    species_list=None,
    species_from_potcar=False,
):
    """
    Routine to read structural static from a POSCAR type file

    Args:
        filename (str): Input filename
        return_velocities (bool): True if the predictor corrector velocities are read (only from MD output)
        species_list (list/numpy.ndarray): A list of the species (if not present in the POSCAR file or a POTCAR in the
        same directory)
        species_from_potcar (bool): True if the species list should be read from the POTCAR file in the same directory

    Returns:
        pyiron.atomistics.structure.atoms.Atoms: The generated structure object

    """
    potcar_file = os.path.join(os.path.dirname(filename), "POTCAR")
    if (species_list is None) and species_from_potcar:
        species_list = get_species_list_from_potcar(potcar_file)
        if len(species_list) == 0:
            warnings.warn("Warning! Unable to read species information from POTCAR")
    file_string = list()
    with open(filename) as f:
        for line in f:
            line = line.strip()
            file_string.append(line)
    return atoms_from_string(
        file_string, read_velocities=return_velocities, species_list=species_list
    )


def get_species_list_from_potcar(filename="POTCAR"):
    """
    Generates the species list from a POTCAR type file

    Args:
        filename (str): Input filename

    Returns:
        list: A list of species symbols

    """
    trigger = "VRHFIN ="
    species_list = list()
    with open(filename) as potcar_file:
        lines = potcar_file.readlines()
        for line in lines:
            line = line.strip()
            if trigger in line:
                str_1 = line.split(trigger)
                str_2 = str_1[-1].split(":")
                species_list.append(str_2[0].replace(" ", ""))
    return species_list


def get_poscar_content(structure, write_species=True, cartesian=True):
    endline = "\n"
    selec_dyn = False
    line_lst = [
        "Poscar file generated with pyiron" + endline,
        "1.0" + endline,
    ]
    for a_i in structure.get_cell():
        x, y, z = a_i
        line_lst.append("{0:.15f} {1:.15f} {2:.15f}".format(x, y, z) + endline)
    atom_numbers = structure.get_number_species_atoms()
    if write_species:
        line_lst.append(" ".join(atom_numbers.keys()) + endline)
    num_str = [str(val) for val in atom_numbers.values()]
    line_lst.append(" ".join(num_str))
    line_lst.append(endline)
    if "selective_dynamics" in structure.get_tags():
        selec_dyn = True
        cartesian = False
        line_lst.append("Selective dynamics" + endline)
    sorted_coords = list()
    selec_dyn_lst = list()
    for species in atom_numbers.keys():
        indices = structure.select_index(species)
        for i in indices:
            if cartesian:
                sorted_coords.append(structure.positions[i])
            else:
                sorted_coords.append(structure.get_scaled_positions()[i])
            if selec_dyn:
                selec_dyn_lst.append(structure.selective_dynamics[i])
    if cartesian:
        line_lst.append("Cartesian" + endline)
    else:
        line_lst.append("Direct" + endline)
    if selec_dyn:
        for i, vec in enumerate(sorted_coords):
            x, y, z = vec
            sd_string = " ".join(["T" if sd else "F" for sd in selec_dyn_lst[i]])
            line_lst.append(
                "{0:.15f} {1:.15f} {2:.15f}".format(x, y, z) + " " + sd_string + endline
            )
    else:
        for i, vec in enumerate(sorted_coords):
            x, y, z = vec
            line_lst.append("{0:.15f} {1:.15f} {2:.15f}".format(x, y, z) + endline)
    return line_lst


def write_poscar(structure, filename="POSCAR", write_species=True, cartesian=True):
    """
    Writes a POSCAR type file from a structure object

    Args:
        structure (pyiron.atomistics.structure.atoms.Atoms): The structure instance to be written to the POSCAR format
        filename (str): Output filename
        write_species (bool): True if the species should be written to the file
        cartesian (bool): True if the positions are written in Cartesian coordinates

    """
    with open(filename, "w") as f:
        f.writelines(
            "".join(
                get_poscar_content(
                    structure=structure,
                    write_species=write_species,
                    cartesian=cartesian,
                )
            )
        )


def atoms_from_string(string, read_velocities=False, species_list=None):
    """
    Routine to convert a string list read from a input/output structure file and convert into Atoms instance

    Args:
        string (list): A list of strings (lines) read from the POSCAR/CONTCAR/CHGCAR/LOCPOT file
        read_velocities (bool): True if the velocities from a CONTCAR file should be read (predictor corrector)
        species_list (list/numpy.ndarray): A list of species of the atoms

    Returns:
        pyiron.atomistics.structure.atoms.Atoms: The required structure object

    """
    string = [s.strip() for s in string]
    string_lower = [s.lower() for s in string]
    atoms_dict = dict()
    atoms_dict["first_line"] = string[0]
    # del string[0]
    atoms_dict["selective_dynamics"] = False
    atoms_dict["relative"] = False
    if "direct" in string_lower or "d" in string_lower:
        atoms_dict["relative"] = True
    atoms_dict["scaling_factor"] = float(string[1])
    unscaled_cell = list()
    for i in [2, 3, 4]:
        vec = list()
        for j in range(3):
            vec.append(float(string[i].split()[j]))
        unscaled_cell.append(vec)
    if atoms_dict["scaling_factor"] > 0.0:
        atoms_dict["cell"] = np.array(unscaled_cell) * atoms_dict["scaling_factor"]
    else:
        atoms_dict["cell"] = np.array(unscaled_cell) * (
            (-atoms_dict["scaling_factor"]) ** (1.0 / 3.0)
        )
    if "selective dynamics" in string_lower:
        atoms_dict["selective_dynamics"] = True
    no_of_species = len(string[5].split())
    species_dict = OrderedDict()
    position_index = 7
    if atoms_dict["selective_dynamics"]:
        position_index += 1
    for i in range(no_of_species):
        species_dict["species_" + str(i)] = dict()
        try:
            species_dict["species_" + str(i)]["count"] = int(string[5].split()[i])
        except ValueError:
            species_dict["species_" + str(i)]["species"] = string[5].split()[i]
            species_dict["species_" + str(i)]["count"] = int(string[6].split()[i])
    atoms_dict["species_dict"] = species_dict
    if "species" in atoms_dict["species_dict"]["species_0"].keys():
        position_index += 1
    positions = list()
    selective_dynamics = list()
    n_atoms = sum(
        [
            atoms_dict["species_dict"][key]["count"]
            for key in atoms_dict["species_dict"].keys()
        ]
    )
    try:
        for i in range(position_index, position_index + n_atoms):
            string_list = np.array(string[i].split())
            positions.append([float(val) for val in string_list[0:3]])
            if atoms_dict["selective_dynamics"]:
                selective_dynamics.append(["T" in val for val in string_list[3:6]])
    except (ValueError, IndexError):
        raise AssertionError(
            "The number of positions given does not match the number of atoms"
        )
    atoms_dict["positions"] = np.array(positions)
    if not atoms_dict["relative"]:
        if atoms_dict["scaling_factor"] > 0.0:
            atoms_dict["positions"] *= atoms_dict["scaling_factor"]
        else:
            atoms_dict["positions"] *= (-atoms_dict["scaling_factor"]) ** (1.0 / 3.0)
    velocities = list()
    try:
        atoms = _dict_to_atoms(atoms_dict, species_list=species_list)
    except ValueError:
        atoms = _dict_to_atoms(atoms_dict, read_from_first_line=True)
    if atoms_dict["selective_dynamics"]:
        selective_dynamics = np.array(selective_dynamics)
        unique_sel_dyn, inverse, counts = np.unique(
            selective_dynamics, axis=0, return_counts=True, return_inverse=True
        )
        count_index = np.argmax(counts)
        atoms.add_tag(selective_dynamics=unique_sel_dyn.tolist()[count_index])
        is_not_majority = np.arange(len(unique_sel_dyn), dtype=int) != count_index
        for i, val in enumerate(unique_sel_dyn):
            if is_not_majority[i]:
                for key in np.argwhere(inverse == i).flatten():
                    atoms.selective_dynamics[int(key)] = val.tolist()
    if read_velocities:
        velocity_index = position_index + n_atoms + 1
        for i in range(velocity_index, velocity_index + n_atoms):
            try:
                velocities.append([float(val) for val in string[i].split()[0:3]])
            except IndexError:
                break
        if not (len(velocities) == n_atoms):
            warnings.warn(
                "The velocities are either not available or they are incomplete/corrupted. Returning empty "
                "list instead",
                UserWarning,
            )
            return atoms, list()
        return atoms, velocities
    else:
        return atoms


def _dict_to_atoms(atoms_dict, species_list=None, read_from_first_line=False):
    """
    Function to convert a generated dict into an structure object

    Args:
        atoms_dict (dict): Dictionary with the details (from string_to_atom)
        species_list (list/numpy.ndarray): List of species
        read_from_first_line (bool): True if we are to read the species information from the first line in the file

    Returns:
        pyiron.atomistics.structure.atoms.Atoms: The required structure object
    """
    is_absolute = not (atoms_dict["relative"])
    positions = atoms_dict["positions"]
    cell = atoms_dict["cell"]
    symbol = str()
    elements = list()
    el_list = list()
    for i, sp_key in enumerate(atoms_dict["species_dict"].keys()):
        if species_list is not None:
            try:
                el_list = np.array([species_list[i]])
                el_list = np.tile(el_list, atoms_dict["species_dict"][sp_key]["count"])
                if isinstance(species_list[i], str):
                    symbol += species_list[i] + str(
                        atoms_dict["species_dict"][sp_key]["count"]
                    )
                else:
                    symbol += species_list[i].Abbreviation + str(
                        atoms_dict["species_dict"][sp_key]["count"]
                    )
            except IndexError:
                raise ValueError(
                    "Number of species in the specified species list does not match that in the file"
                )
        elif "species" in atoms_dict["species_dict"][sp_key].keys():
            el_list = np.array([atoms_dict["species_dict"][sp_key]["species"]])
            el_list = np.tile(el_list, atoms_dict["species_dict"][sp_key]["count"])
            symbol += atoms_dict["species_dict"][sp_key]["species"]
            symbol += str(atoms_dict["species_dict"][sp_key]["count"])
        elif read_from_first_line:
            if not (
                len(atoms_dict["first_line"].split())
                == len(atoms_dict["species_dict"].keys())
            ):
                raise AssertionError()
            el_list = np.array(atoms_dict["first_line"].split()[i])
            el_list = np.tile(el_list, atoms_dict["species_dict"][sp_key]["count"])
            symbol += atoms_dict["first_line"].split()[i]
            symbol += str(atoms_dict["species_dict"][sp_key]["count"])
        elif species_list is None:
            raise ValueError(
                "Species list should be provided since pyiron can't detect species information"
            )
        elements.append(el_list)
    elements_new = list()
    for ele in elements:
        for e in ele:
            elements_new.append(e)
    elements = elements_new
    if is_absolute:
        atoms = Atoms(elements, positions=positions, cell=cell, pbc=True)
    else:
        atoms = Atoms(elements, scaled_positions=positions, cell=cell, pbc=True)
    return atoms


def vasp_sorter(structure):
    """
    Routine to sort the indices of a structure as it would be when written to a POSCAR file

    Args:
        structure (pyiron.atomistics.structure.atoms.Atoms): The structure whose indices need to be sorted

    Returns:
        list: A list of indices which is sorted by the corresponding species for writing to POSCAR

    """
    atom_numbers = structure.get_number_species_atoms()
    sorted_indices = list()
    for species in atom_numbers.keys():
        indices = structure.select_index(species)
        for i in indices:
            sorted_indices.append(i)
    return np.array(sorted_indices)


def manip_contcar(filename, new_filename, add_pos):
    """
    Manipulate a CONTCAR/POSCAR file by adding something to the positions

    Args:
        filename (str):  Filename/path of the input file
        new_filename (str): Filename/path of the output file
        add_pos (list/numpy.ndarray): Array of values to be added to the positions of the input

    """
    actual_struct = read_atoms(filename)
    n = 0
    direct = True
    with open(filename, "r") as f:
        lines = f.readlines()
        for line in lines:
            if "Direct" in line or "Cartesian" in line:
                direct = "Direct" in line
                break
            n += 1
    pos_list = list()
    sd_list = list()
    if len(lines[n + 1].split()) == 6:
        for line in lines[n + 1 : n + 1 + len(actual_struct)]:
            pos_list.append([float(val) for val in line.split()[0:3]])
            sd_list.append(["T" in val for val in line.split()[3:]])
    else:
        for line in lines[n + 1 : n + 1 + len(actual_struct)]:
            pos_list.append([float(val) for val in line.split()[0:3]])
    old_pos = np.array(pos_list)
    if direct:
        add_pos_rel = np.dot(add_pos, np.linalg.inv(actual_struct.cell))
        new_pos = old_pos + add_pos_rel
    else:
        new_pos = old_pos + add_pos
    new_lines = lines
    new_pos_str = np.array(new_pos, dtype=str)
    if len(sd_list) > 0:
        bool_list = np.zeros_like(old_pos, dtype=str)
        bool_list[:] = "F"
        bool_list[np.array(sd_list)] = "T"
        for i, pos in enumerate(new_pos_str):
            linestr = np.append(pos, bool_list[i])
            new_lines[n + 1 + i] = " ".join([str(val) for val in linestr]) + "\n"
    else:
        for i, pos in enumerate(new_pos_str):
            linestr = pos
            new_lines[n + 1 + i] = " ".join([str(val) for val in linestr]) + "\n"

    # Exclude predictor corrector positions
    if len(new_lines[n + len(new_pos_str) :]) >= 2 * len(new_pos_str):
        new_lines = new_lines[: n + 2 * (len(new_pos_str)) + 2]

    with open(new_filename, "w") as f:
        for new_line in new_lines:
            f.write(new_line)
